function SQLiteHandler() {
  // Get database storage service
  this.storageService = Components.classes["@mozilla.org/storage/service;1"]
      .getService(Components.interfaces.mozIStorageService);
};

SQLiteHandler.prototype = {
  dataConnection: null,   // Connection to current database (mozIStorageConnection)
  aTableData: null,       // Stores 2D array of table data
	aTableType: null,
  aColumns: null,

	colNameArray: null,
	resultsArray: null,
	statsArray: null,

	lastErrorString: "",
	sErrorMessage: null,
	sExceptionName: null,
	sExceptionMessage: null,
				
  // ConnectToDatabase: Connects and sets dataConnection
  ConnectToDatabase: function(nsIFile)
  {   
    this.closeConnection();

		try
		{
      this.dataConnection = this.storageService.openDatabase(nsIFile);
			// if the db does not exist it does not give us any indication
			// this.dataConnection.lastErrorString returns "not an error"
		}
		catch (e)
		{
			this.onSqlError(e, "Error in opening file " + nsIFile.leafName + 
											   " - perhaps, this is not an sqlite db file", null);
		}
		
		if(this.dataConnection == null)
			return false;
		
		return true;
  },

  //closeConnection: Free the database resources (right now this doesn't do much)
  closeConnection: function() 
  {
    if (this.dataConnection != null)
	    this.dataConnection = null; //close() doesn't work, so C++ style  
	    //this.dataConnection.close();  
    
    this.aTableData = null;
    this.aTableType = null;
    this.aColumns = null;
  },

	getFileName: function()
	{
    if (this.dataConnection != null)
	    return this.dataConnection.databaseFile.leafName;
	  return null;
	},

	getSqliteVersion: function()
	{
		var sQuery = "SELECT sqlite_version()";
		this.selectQuery(sQuery);

		var sVersion = this.aTableData[0][0];
		return sVersion;
	},
	
	getSchemaVersion: function()
	{
    return this.dataConnection.schemaVersion;
	},
	
	setSetting: function(sSetting, sValue)
	{
		if (sSetting == "encoding" || sSetting == "temp_store_directory")
			sValue = "'" + sValue + "'";
		var sQuery = "PRAGMA " + sSetting + " = " + sValue;
		//do not execute in a transaction; some settings will cause error
		this.selectQuery(sQuery);
		
		return this.getSetting(sSetting);
	},
	
	getSetting: function(sSetting)
	{
		var iValue = null;
		var sQuery = "PRAGMA " + sSetting;
		this.selectQuery(sQuery);

		iValue = this.aTableData[0][0];
		return iValue;
	},
	
	getMasterInfo: function(sObjName, sField)
	{
		var sQuery = "SELECT " + sField + " FROM sqlite_master WHERE name='" + sObjName + "'";
		this.selectQuery(sQuery);
		return this.aTableData[0][0];
	},
	
  // getObjectList: Retuns an array of names of type=argument 
	// in the current database. Type = table|index|view|trigger
  getObjectList: function(sObjectType) 
  {
  	return this.getSchemaArray(sObjectType);
  },
  
  //getSchemaArray: must return an array, empty array if no object found
  //calling functions can evaluate the returned value's length property
  //and if it is 0, it means corresponding objects do not exist in the database
  getSchemaArray: function(sType)
  {
    var aResult = new Array();
    var iCount = 0;
    
    var sQuery = "SELECT name FROM sqlite_master WHERE type = '"
					+ sType + "' ORDER BY name";
		this.selectQuery(sQuery);

		for (var i = 0; i < this.aTableData.length; i++)
			aResult.push(this.aTableData[i][0]);

		return aResult;
  },

  // loadTableData: retrieves data from a table including rowid if needed
  loadTableData: function(sObjType, sObjName, sCondition, iLimit, iOffset)
  {
  	if (sObjType != "table" && sObjType != "view")
  		return false;
  		
  	iLimit = parseInt(iLimit);
  	if (isNaN(iLimit))
  		iLimit = -1;
  		
  	iOffset = parseInt(iOffset);
  	if (isNaN(iOffset))
  		iOffset = 0;
  		
		var extracol = "";
  	var sLimitClause = " LIMIT " + iLimit + " OFFSET " + iOffset;
  	
  	if (sObjType == "table")
  	{
	    //find whether the rowid is needed 
			//or the table has an integer primary key
			var rowidcol = this.getTableRowidCol(sObjName);
			if (rowidcol["name"] == "rowid")
				extracol = " `rowid`, ";
		}
    //table having columns called rowid behave erratically
		this.selectQuery("SELECT " + extracol + " * FROM `" + sObjName + "` "
						+ sCondition + sLimitClause);
		return true;
	},

	//for tables and views
	getTableCount: function(sObjName, sCondition)
	{
		var iValue = 0;
		var sQuery = "SELECT count(*) FROM `" + sObjName + "` " + sCondition;
		this.selectQuery(sQuery);

		iValue = this.aTableData[0][0];
		return iValue;
	},
	
  // getRecords: return aTableData 
  getRecords: function()
  {
    return this.aTableData;         
	},

  getRecordTypes: function()
  {
    return this.aTableType;         
	},

  getColumns: function()
  {
    return this.aColumns;         
	},
		
  // selectQuery : execute a select query and store the results
  selectQuery: function(sQuery)
  {
  	//alert(sQuery);
    this.aTableData = new Array();
    this.aTableType = new Array();
// if aColumns is not null, there is a problem in tree display
//	this.aColumns = new Array();
    this.aColumns = null;        
    var bResult = false;
 
    try {		// mozIStorageStatement
			var stmt = this.dataConnection.createStatement(sQuery);
		}
		catch (e) {
			// statement will be undefined because it throws error);
			this.onSqlError(e, "Likely SQL syntax error: " + sQuery, 
			  		this.dataConnection.lastErrorString);
			this.setErrorString();
			return false;
		}
		
    var iCols = 0;
    var iType, colName;
		try	{
      // do not use stmt.columnCount in the for loop, fetches the value again and again
      iCols = stmt.columnCount;
			this.aColumns = new Array();
			var aTemp, aType;
			for (var i = 0; i < iCols; i++)	{
        colName = stmt.getColumnName(i);
        aTemp = [colName, iType];
        this.aColumns.push(aTemp);  
			}
		} catch (e) { 
			this.onSqlError(e, "Error while fetching column name: " + colName, null);
    	this.setErrorString();
    	return false;
  	}

    var cell;
   	var bFirstRow = true;
    try {
      while (stmt.executeStep()) 
			{
        aTemp = [];//new Array();
        aType = [];
        for (i = 0; i < iCols; i++) 
  			{
          iType = stmt.getTypeOfIndex(i);//0=null, 1=int, 2=float, 3=string, 4=blob
	        if (bFirstRow)
	        {
	        	this.aColumns[i][1] = iType;
	        }
          switch (iType) {
            case stmt.VALUE_TYPE_NULL: 
							cell = g_strForNull;
							break;
            case stmt.VALUE_TYPE_INTEGER:
            //use getInt64, not getInt32 otherwise long int
            // as in places.sqlite/cookies.sqlite shows funny values
							cell = stmt.getInt64(i);
							break;
            case stmt.VALUE_TYPE_FLOAT:
							cell = stmt.getDouble(i);
							break;
            case stmt.VALUE_TYPE_TEXT:
							cell = stmt.getString(i);
							break;
            case stmt.VALUE_TYPE_BLOB: //todo - handle blob properly
							cell = stmt.getString(i);
							break;
            default: sData = "<unknown>"; 
          }
          aTemp.push(cell);
          aType.push(iType);
        }
        this.aTableData.push(aTemp);
        this.aTableType.push(aType);
        bFirstRow = false;
      }
    } catch (e) { 
			this.onSqlError(e, "Query: " + sQuery + " - executeStep failed", null);
    	this.setErrorString();
    	return false;
    } finally {
    	//must be reset after executeStep
    	stmt.reset();
    }
		this.setErrorString();
    return true;
  },

  // getTableRowidCol : execute a pragma query and return the results
  getTableRowidCol: function(sTableName)
  {
		var sQuery = "PRAGMA table_info(`" + sTableName + "`)";
		this.selectQuery(sQuery);
		var aReturn = [];
		var iPk, iType, iName, iCid;
		for (var i in this.aColumns)
		{
			switch (this.aColumns[i][0])
			{
				case "pk": iPk = i; break;
				case "type": iType = i; break;
				case "name": iName = i; break;
				case "cid": iCid = i; break;
			}
		}
		var iNumPk = 0;
		for(var i = 0; i < this.aTableData.length; i++)
		{
			var row = this.aTableData[i];
			var type = row[iType];
			var pk = row[iPk];
			type = type.toUpperCase();
			if(pk == 1 && type == "INTEGER")
			{
				iNumPk++;
				var name = row[iName];
				var cid = row[iCid];
				aReturn["name"] = name;
				aReturn["cid"] = cid;
			}
		}
		if (iNumPk == 1)
			return aReturn;
		
		aReturn["name"] = "rowid";
		aReturn["cid"] = 0;
		return aReturn;
  },
    
  // getTableColumns : execute a pragma query and return the results
  getTableColumns: function(sTableName)
  {
		var sQuery = "PRAGMA table_info(`" + sTableName + "`)";
		this.selectQuery(sQuery);
		
		var aCols = new Array();
		for (var i in this.aColumns)
		{
			var aTemp = [i, this.aColumns[i][1]];
			aCols[this.aColumns[i][0]] = aTemp;
		}
		var aResult = [this.aTableData, aCols, this.aTableType];
		return aResult;
  },

	getTableColumnsWithDefaultValue: function(sTableName)
	{
		var aResult = [];
		var info = this.getTableColumns(sTableName);
		var columns = info[0];
		for(var i = 0; i < columns.length; i++)
		{
			if (columns[i][info[1]["dflt_value"][0]] != g_strForNull)
				aResult.push(columns[i][info[1]["name"][0]]);
		}
		return aResult;
	},
	
  // getTableInfo : execute a pragma query and return the results
  getTableInfo: function(sTableName,ciInfoType)
  {
		var aResult = [];
		var sQuery = "PRAGMA table_info(`" + sTableName + "`)";
		this.selectQuery(sQuery);
		aResult["numFields"] = this.aTableData.length;

		sQuery = "PRAGMA index_list(`" + sTableName + "`)";
		this.selectQuery(sQuery);
		if(this.aTableData != null)
			aResult["numIndexes"] = this.aTableData.length;
		else
			aResult["numIndexes"] = 0;
		
		aResult["numRecords"] = this.getTableCount(sTableName, "");

   	return aResult;
  },
  
  // getIndexInfo : execute a pragma query and return the results
  getIndexInfo: function(sIndexName)
  {
		//to find the table and sql of the index
		var sQuery = "SELECT sql, tbl_name FROM sqlite_master WHERE type='index' and name='" + sIndexName + "'";
		this.selectQuery(sQuery);

		var aReturn = [];
		aReturn["sql"] = this.aTableData[0][0];
		aReturn["table"] = this.aTableData[0][1];

		//to find fields in the index
		var sQuery = "PRAGMA index_info(`" + sIndexName + "`)";
		this.selectQuery(sQuery);

   	var iRequiredField = 2;
		var cols = [];
		for(var i = 0; i < this.aTableData.length; i++)
		{
			cols.push(this.aTableData[i][iRequiredField]);
		}
		aReturn["cols"] = cols;

		//to find whether duplicates allowed
		aReturn["unique"] = 0;
		var sQuery = "PRAGMA index_list(`" + aReturn["table"] + "`)";
		this.selectQuery(sQuery);
		for(var i = 0; i < this.aTableData.length; i++)
		{
			if(this.aTableData[i][1] == sIndexName)
				aReturn["unique"] = this.aTableData[i][2];
		}
		
		return aReturn;
  },
    
	select : function(file,sql,param)
	{
		var ourTransaction = false;
		if (this.dataConnection.transactionInProgress)
		{
			ourTransaction = true;
			this.dataConnection.beginTransactionAs(this.dataConnection.TRANSACTION_DEFERRED);
		}
		var statement = this.dataConnection.createStatement(sql);
    if (param)
		{
			for (var m = 2, arg = null; arg = arguments[m]; m++) 
			{
				statement.bindUTF8StringParameter(m-2, arg);
			}
		}
		try
		{
			var dataset = [];
			while (statement.executeStep())
			{
				var row = [];
				for (var i = 0, k = statement.columnCount; i < k; i++)
				{
					row[statement.getColumnName(i)] = statement.getUTF8String(i);
				}
				dataset.push(row);
			}
			// return dataset;	
		}
		finally 
		{
			statement.reset();
		}
		if (ourTransaction)
		{
			this.dataConnection.commitTransaction();
		}
        return dataset;	
	},
	
	executeQueryWithParams: function(sQuery, param)
	{
		//commit, if some leftover transaction is in progress
		if (this.dataConnection.transactionInProgress)
			this.dataConnection.commitTransaction();

		//begin a transaction, iff no transaction in progress
		if (!this.dataConnection.transactionInProgress)
			this.dataConnection.beginTransaction();
		
    try 
		{
			// mozIStorageStatement
			var statement = this.dataConnection.createStatement(sQuery);
		}
		catch(e)
		{
			// statement will be undefined because it throws error);
			this.onSqlError(e, "Likely SQL syntax error: " + sQuery, 
							this.dataConnection.lastErrorString);
			this.setErrorString();
			return false;
		}

		if (param)
		{
			for (var m = 1, arg = null; arg = arguments[m]; m++) 
			{
				statement.bindUTF8StringParameter(m - 1, arg);
			}
		}
		try{
			statement.execute();
			this.setErrorString();
		}
		catch(e) {
			// statement will be undefined because it throws error);
			this.onSqlError(e, "Execute failed: " + sQuery, 
							this.dataConnection.lastErrorString);
			this.setErrorString();
			return false;
		}
		finally {
			statement.reset();
		}
		//commit transaction, if reached here
		if (this.dataConnection.transactionInProgress)
			this.dataConnection.commitTransaction();

		return true;
	},	

	executeMultiple: function(sQuery)
	{
		//commit, if some leftover transaction is in progress
		if (this.dataConnection.transactionInProgress)
			this.dataConnection.commitTransaction();

		//begin a transaction, iff no transaction in progress
		if (!this.dataConnection.transactionInProgress)
			this.dataConnection.beginTransaction();

		var aQueries = [];
		var arr = sQuery.split(";");
		var i = 0;
		var str = arr[i];
		while(true)
		{
			var statement = this.isValidStatement(str);
			if (statement)
			{
				aQueries.push(str);
				alert(aQueries.length + " = " + str);
				try {
				alert("executing ");
					statement.execute();
					this.setErrorString();
				}
				catch (e)	{
					// statement will be undefined because it throws error);
					this.onSqlError(e, "Execute failed: " + str, 
									this.dataConnection.lastErrorString);
					this.setErrorString();
					if (this.dataConnection.transactionInProgress) {
						this.dataConnection.rollbackTransaction();
					}
					return false;
				}
				finally {
					//statement.reset();
				}
				i++;
				if (i >= arr.length)
				{
								alert("i > arr.length ");
					this.dataConnection.commitTransaction();
					return true;
				}
				else
					str = arr[i];
			}
			else {
				// statement creation failed. append the next portion and check for validity
				i++;
				if (i >= arr.length)
				{
					this.onSqlError(e, "Likely SQL syntax error: " + arr[i], 
								this.dataConnection.lastErrorString);
					this.setErrorString();
					return false;
				}
				else
				{
					str = str + ";" + arr[i];
								alert("joined  " + str);
					continue;
				}
			}
		}
		//commit transaction, if reached here
		if (this.dataConnection.transactionInProgress)
			this.dataConnection.commitTransaction();

		return true;
	},
	
	isValidStatement: function(sQuery)
	{
    try {
			var statement = this.dataConnection.createStatement(sQuery);
		}
		catch (e) {
			return false;
		}
		return statement;
	},
	
	executeTransaction: function(aQueries)
	{
		//commit, if some leftover transaction is in progress
		if (this.dataConnection.transactionInProgress)
			this.dataConnection.commitTransaction();

		//begin a transaction, iff no transaction in progress
		if (!this.dataConnection.transactionInProgress)
			this.dataConnection.beginTransaction();

		for(var i = 0; i < aQueries.length; i++)
		{
	    try {
				var statement = this.dataConnection.createStatement(aQueries[i]);
			}
			catch (e) {
				// statement will be undefined because it throws error);
				this.onSqlError(e, "Likely SQL syntax error: " + aQueries[i], 
								this.dataConnection.lastErrorString);
				this.setErrorString();
				if (this.dataConnection.transactionInProgress) {
					this.dataConnection.rollbackTransaction();
				}
				return false;
			}
	
			try {
				statement.execute();
				this.setErrorString();
			}
			catch (e)	{
				// statement will be undefined because it throws error);
				this.onSqlError(e, "Execute failed: " + aQueries[i], 
								this.dataConnection.lastErrorString);
				this.setErrorString();
				if (this.dataConnection.transactionInProgress) {
					this.dataConnection.rollbackTransaction();
				}
				return false;
			}
			finally {
				statement.reset();
			}
		}
		//commit transaction, if reached here
		if (this.dataConnection.transactionInProgress)
			this.dataConnection.commitTransaction();

		return true;
	},	

	getLastError: function()
	{
		return this.lastErrorString;
	},
	
	setErrorString: function()
	{
		this.lastErrorString = this.dataConnection.lastErrorString;
	},
	
	onSqlError: function(ex, msg, SQLmsg)
	{
		try
		{
			this.sExceptionName = ex.name;
			this.sExceptionMessage = ex.message;
			this.sErrorMessage = msg;
			if (SQLmsg != null)
				this.sErrorMessage += " [ " + SQLmsg + " ]";
		}
		catch(e)
		{
			alert("Error inside the function onSqlError; exception = " + e.name + "; " + e.message);
			return false;
		}
		alert(this.sErrorMessage + "\n" +
		      "Exception Name: " + this.sExceptionName + "\n" +
					"Exception Message: " + this.sExceptionMessage);
		return true;
	}
};
