// SQLiteManager extension
var Database = new SQLiteHandler();
var smExtManager = null;

var SQLiteManager = {    
  prefs: null,
  
  //for display in the browse tree
  miLimit: -1,
  miOffset: 0,
  miCount: 0,
  sBrowseObjName: null,
  msBrowseCondition: null,
  
  //whether to implicitly quote the values or not
  mbImplicitlyQuote: true,
  
  // Currently selected database file (nsIFile)
  sCurrentDatabase: null, 
  
  //an array containing names of current table, index, view and trigger
  aCurrObjNames: [],   
  
  //to store the latest selection in tree showing db objects
  mostCurrObjName: null,
  mostCurrObjType: null,
  
  mbAfterStartup: true,
  // an array of 4 arrays; 
  // each holding names of tables, indexes, views and triggers
  aObjNames: [],
  
  aObjTypes: ["table","view","index","trigger"],

  iCurrentRow: -1,
  clipService: null,      // Clipboard service: nsIClipboardHelper

  // Status bar: panels for displaying various info
  sbPanel: [],         

  iDebugMode: 0,          // Debug mode

  //store extension pertaining mgmt info externally/internally
  m_bMgmtDataIsInDb: true,

  experiment: function()
  {

  },

  // Startup: called ONCE during the browser window "load" event
  Startup: function() 
  {
    this.mbAfterStartup = true;
    //create the menus by associating appropriate popups
    this.createMenu();

    g_smBundle = document.getElementById("sm-bundle");

    //enable pragma version options for gecko 1.9
    if (gbGecko_1_9_b3)
      document.getElementById("gb-pragmaVersion").hidden = false;

    // Load clipboard service
    this.clipService= Components
    	.classes["@mozilla.org/widget/clipboardhelper;1"]
       .getService(Components.interfaces.nsIClipboardHelper);

    //get the nodes for Status bar panels
    var panels = ["display","table","view","index","trigger"];
    for(var i = 0; i < panels.length; i++)
    {
      this.sbPanel[panels[i]] = document.getElementById("sbPanel-" + panels[i]);
    }
    //global var in misc.js for sbpanel[display]
    gSbPanelDisplay = this.sbPanel["display"];

    //preferences service to add oberver
    // and then observe changes via the observe function
    //see http://developer.mozilla.org/en/docs/Adding_preferences_to_an_extension
    //initialize the preference service with the correct branch
    this.prefs = sm_prefsBranch;
    //query interface to be able to use addObserver method
    this.prefs.QueryInterface(Components.interfaces.nsIPrefBranch2);
    //now, add the observer which will be implemented using observe method
    //calling removeObserver when done with observing helps the memory
    this.prefs.addObserver("", this, false);

    //manage changes when the following preferences change.
    //although the this.observe function handles the observer
    //calling these functions reduces the size of the code
    //It is like fooling observe into believing that the following
    //preferences have changed.
    this.observe("", "nsPref:changed", "hideMainToolbar");
    this.observe("", "nsPref:changed", "showMainToolbarDatabase");
    this.observe("", "nsPref:changed", "showMainToolbarTable");
    this.observe("", "nsPref:changed", "showMainToolbarIndex");
    this.observe("", "nsPref:changed", "sqliteFileExtensions");
    this.observe("", "nsPref:changed", "displayNumRecords");
    this.observe("", "nsPref:changed", "displayNumRecords");
    this.observe("", "nsPref:changed", "sqlTextboxRows");

    //initialize the structure tree and tabpanel
    this.loadDBStructure();

    // opening with last used DB if preferences set to do so
    var bPrefVal = sm_getPreferenceValue("openWithLastDb", false);
		if(bPrefVal)
		{
			var sPrefVal = sm_getPreferenceValue("mruPath.1", "");
			if(sPrefVal.length > 0)
			{
				//Last used DB found, open this DB
				var newfile = Components.classes["@mozilla.org/file/local;1"]
									.createInstance(Components.interfaces.nsILocalFile);
				newfile.initWithPath(sPrefVal);
				//if the last used file is not found, bail out
				if(!newfile.exists())
				{
					alert(sm_getLStr("lastDbDoesNotExist") + sPrefVal);
				  return;
				}
				bPrefVal = sm_getPreferenceValue("promptForLastDb", true);
				if(bPrefVal)
				{
					var prompts = Components
									.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);							
					var check = {value: false}; // default the checkbox to false
					var result = prompts.confirmCheck(null, sm_getLStr("promptLastDbTitle"),
					sm_getLStr("promptLastDbAsk")+ "\n" + sPrefVal + "?",
          sm_getLStr("promptLastDbOpen"), check);

					if(!result)
						return;
					//update the promptForLastDb preference
					bPrefVal = sm_setPreferenceValue("promptForLastDb", !check.value);
				}
				//assign the new file (nsIFile) to the current database
				this.sCurrentDatabase = newfile;
				this.setDatabase(this.sCurrentDatabase);
			}
		}
		//load the previously opened tab
    this.loadTabWithId(this.getSelectedTabId());
		return;
  },

  // Shutdown: called ONCE during the browser window "unload" event
  Shutdown: function()
  {
    //Destruction - this should be fired once you're done observing
    //Failure to do so may result in memory leaks.	
    this.prefs.removeObserver("", this);
    this.prefService= null;

    this.clipService= null;
  },  

	createMenu: function()
	{
		var suffixes = ["table", "index", "view", "trigger"];
		for(var i = 0; i < suffixes.length; i++)
		{
		  var suffix = suffixes[i];
		  var mp = document.getElementById("menupopup-" + suffix);
		  var clone = mp.cloneNode(true);
		  clone.setAttribute("id", "menupopup-main-" + suffix);
		  var menu = document.getElementById("menu-" + suffix);
		  ClearElement(menu);
		  menu.appendChild(clone);
		}
	},

	changeDbSetting: function(sSetting)
	{
		var node = document.getElementById("pr-" + sSetting);
		var sVal = node.value;
		var newVal = Database.setSetting(sSetting, sVal);
		node.value = newVal;
	},
		
	setTreeTableContextMenu: function(event)
	{
	
	},
	
	setTreeStructureContextMenu: function(event)
	{
		document.getElementById("treechildren-db-structure")
		  .setAttribute("context","");
		var tree = document.getElementById("tree-db-structure");
		var idx = tree.currentIndex;
		// idx = -1 if nothing is selected; says xulplanet element reference
		if(idx == -1)
			idx = 0;
		var objName = tree.view.getCellText(idx, tree.columns.getColumnAt(0));
		var level = tree.view.getLevel(idx);

		//there is a database object at level 1 only
		if (level == 0)
		{
			if (objName.indexOf("Tables") == 0)
				document.getElementById("treechildren-db-structure")
				  .setAttribute("context","menupopup-create-table");
			if (objName.indexOf("Indexes") == 0)
				document.getElementById("treechildren-db-structure")
				  .setAttribute("context","menupopup-create-index");
			if (objName.indexOf("Views") == 0)
				document.getElementById("treechildren-db-structure")
				  .setAttribute("context","menupopup-create-view");
			if (objName.indexOf("Triggers") == 0)
				document.getElementById("treechildren-db-structure")
				  .setAttribute("context","menupopup-create-trigger");
			return;
		}	
		if (level != 1)
			return;
		
		var r_type = Database.getMasterInfo(objName, "type");
		// context menu to be based on r_type
		document.getElementById("treechildren-db-structure")
		  .setAttribute("context","menupopup-" + r_type);
		//context menu assignment ends
	},
	
	observe: function(subject, topic, data)
	{
		if (topic != "nsPref:changed")
			return;
		
		switch(data)
		{
			case "hideMainToolbar":
				var bPrefVal = sm_getPreferenceValue("hideMainToolbar", false);
				this.setElementDisplay("hbox-main-toolbar", !bPrefVal);
				break;
			case "showMainToolbarDatabase":
				var bPrefVal = sm_getPreferenceValue("showMainToolbarDatabase", true);
				this.setElementDisplay("sm-toolbar-database", bPrefVal);
				break;
			case "showMainToolbarTable":
				var bPrefVal = sm_getPreferenceValue("showMainToolbarTable", true);
				this.setElementDisplay("sm-toolbar-table", bPrefVal);
				break;
			case "showMainToolbarIndex":
				var bPrefVal = sm_getPreferenceValue("showMainToolbarIndex", true);
				this.setElementDisplay("sm-toolbar-index", bPrefVal);
				break;
			case "sqliteFileExtensions":
				var sExt = sm_getPreferenceValue("sqliteFileExtensions", "sqlite");
        // Load profile folder's sqlite db files list into dropdown 
				this.populateProfileDBList(sExt);   
				break;
			case "tempSearchQuery":
				var sPrefVal = sm_getPreferenceValue("tempSearchQuery", "");
				this.msBrowseCondition = sPrefVal;
				this.loadTabBrowse();
				break;
			case "displayNumRecords":
				var iPrefVal = sm_getPreferenceValue("displayNumRecords", -1);
				this.miLimit = iPrefVal;
				this.miOffset = 0;
				break;
			case "sqlTextboxRows":
				var iPrefVal = sm_getPreferenceValue("sqlTextboxRows", 2);
				document.getElementById("txtSqlStatement").setAttribute("rows", iPrefVal);
				break;
			case "implicitlyQuote":
				var bPrefVal = sm_getPreferenceValue("implicitlyQuote", true);
				this.mbImplicitlyQuote = bPrefVal;
				break;
		}
	},
		 
	setElementDisplay: function(elementId, bDisplay)
	{
		var elt = document.getElementById(elementId);
		elt.hidden = !bDisplay;
	},

  refresh: function()
  {
    if (this.sCurrentDatabase == null)
    	return false;
    this.loadDBStructure();
    return true; 
  },

  //loadDBStructure: populates the schematree based on selected database
  //must be called whenever a database is opened/closed
  //and whenever the schema changes
  loadDBStructure: function()
  {
    var tree = document.getElementById("tree-db-structure");
    // if no database is selected
    // draw an empty tree, correct status bar info
    if (this.sCurrentDatabase == null)
    {
		TreeViewDb.removeChildData();
		tree.view = TreeViewDb;

		for(var i = 0; i < this.aObjTypes.length; i++)
		{
			var type = this.aObjTypes[i];
			this.aCurrObjNames[type] = null;
			this.sbPanel[type].label = this.aCurrObjNames[type];
		}
		return;
	}

		//requery for all the objects afresh and redraw the tree
		this.aObjNames["table"] = Database.getObjectList("table");
		this.aObjNames["index"] = Database.getObjectList("index");
		this.aObjNames["view"] = Database.getObjectList("view");
		this.aObjNames["trigger"] = Database.getObjectList("trigger");
		
		TreeViewDb.setChildData(this.aObjNames);
		tree.view = TreeViewDb;

		//now assign the current objects
		for(var i = 0; i < this.aObjTypes.length; i++)
		{
			var type = this.aObjTypes[i];
			if(this.aObjNames[type].length > 0)
			{
				var bFound = false;
				if(this.aCurrObjNames[type])
				{
					for(var n = 0; n < this.aObjNames[type].length; n++)
					{
						if(this.aCurrObjNames[type] == this.aObjNames[type][n])
						{
							bFound = true;
							break;
						}
					}
				}
				if(!bFound)
					this.aCurrObjNames[type] = this.aObjNames[type][0];
			}
			else
				this.aCurrObjNames[type] = null;
				
			this.sbPanel[type].label = this.aCurrObjNames[type];
		}
		this.getDbObjectInfo();
  },
		 		 
	//getDbObjectInfo: this function must show the structural info about the
	// selected database object (table, index, view & trigger)
  getDbObjectInfo: function()
  {
		var tree = document.getElementById("tree-db-structure");
		//if this function is being called first time after Startup
		//then assign the persistent values
		var idx;
		if (this.mbAfterStartup)
		{
			this.mbAfterStartup = false;
			idx = document.getElementById("lbl-db-currIndex").value;
			tree.view.selection.select(idx);
		}
		else
			idx = tree.currentIndex;

		document.getElementById("lbl-db-currIndex").setAttribute("value",idx);
		// idx = -1 if nothing is selected; says xulplanet element reference
		if(idx == -1)
			idx = 0;
		var objName = tree.view.getCellText(idx, tree.columns.getColumnAt(0));
//			var objType = tree.view.getCellText(tree.view.getParentIndex(idx), tree.columns.getColumnAt(0));
		var level = tree.view.getLevel(idx);

		//there is a database object at level 1 only
		if(level != 1)
		{
			this.mostCurrObjName = null;
			this.mostCurrObjType = null;
			return false;
		}	
		
		//query the type and name of this object
		var r_type = Database.getMasterInfo(objName, "type");
		var r_name = Database.getMasterInfo(objName, "name");
			
		//assign current selection in tree as current object
		this.aCurrObjNames[r_type] = r_name;
		//show the current object in the statusbarpanel too
		this.sbPanel[r_type].label = r_name;

		this.mostCurrObjName = r_name;
		this.mostCurrObjType = r_type;

		this.loadTabStructure();
		this.loadTabBrowse();

		return true;
  },

  loadTabStructure: function()
  {
    //no need to waste resources if this tab is not selected
    if(this.getSelectedTabId() != "tab-structure")
    	return false;
    	
    if (this.sCurrentDatabase == null)
    	return false;

		//hide the hboxes containing object specific operation buttons
		//later enable one appropriate hbox according to the selection in tree
		for(var i=0; i < this.aObjTypes.length; i++)
		{
			document.getElementById("gb-master-ops-" + this.aObjTypes[i])
			  .hidden = true;
			document.getElementById("gb-more-info-" + this.aObjTypes[i])
			  .hidden = true;
		}			

		//there is a database object at level 1 only
		if(this.mostCurrObjName == null)
		{
			document.getElementById("gb-master-info").hidden = true;
			return false;
		}	
		
		//display this groupbox only if it will show something meaningful
		document.getElementById("gb-master-info").hidden = false;
		
		var r_name = this.mostCurrObjName;
		var r_type = this.mostCurrObjType;
		document.getElementById("gb-master-ops-" + r_type).hidden = false;

		var r_tbl_name = Database.getMasterInfo(r_name, "tbl_name");
		var r_rootpage = Database.getMasterInfo(r_name, "rootpage");
		var r_sql = Database.getMasterInfo(r_name, "sql");
			
		var row = {"type":r_type, "name": r_name, "tbl_name": r_tbl_name,"rootpage":r_rootpage, "sql":r_sql};
		var fld = ["type","name","tbl_name","rootpage", "sql"];
		var node;
		for(var i=0; i < fld.length; i++)
		{
			node = document.getElementById("str-"+fld[i]);
			var val = row[fld[i]];
			if(fld[i] == "type")
			{
				val = val.toUpperCase();
			}
			node.value = val;
			node.setAttribute("style", "font-weight:bold;font-size:12px;");
		}		

		//do the following for table/view
		if(r_type == "table")
		{
			var box = document.getElementById("gb-more-info-" + r_type);
			box.hidden = false;
			this.printTableInfo(this.aCurrObjNames[r_type]);
		}
		if(r_type == "index")
		{
			var box = document.getElementById("gb-more-info-" + r_type);
			box.hidden = false;
			this.printIndexInfo(this.aCurrObjNames[r_type]);
		}
		return true;
  },

  printTableInfo: function(sTable)
  {
		var info = Database.getTableColumns(sTable);
		var vbox = document.getElementById("vb-more-info-table-columns");
		ClearElement(vbox);

		var cols = info[0];
		for(var i = 0; i < cols.length; i++)
		{
			var hbox = document.createElement("hbox");
			var tb = document.createElement("textbox");
			tb.setAttribute("style","width:12ex");
			tb.setAttribute("readonly","true");
			tb.setAttribute("value", cols[i][info[1]["name"][0]]);
			hbox.appendChild(tb);

			var tb = document.createElement("textbox");
			tb.setAttribute("style","width:10ex");
			tb.setAttribute("readonly","true");
			tb.setAttribute("value", cols[i][info[1]["type"][0]]);
			hbox.appendChild(tb);

			var tb = document.createElement("checkbox");
			tb.setAttribute("style","width:7ex");
			tb.setAttribute("disabled","true");
			if (cols[i][info[1]["pk"][0]] == 1)
				tb.setAttribute("checked", true);
			hbox.appendChild(tb);

			var tb = document.createElement("checkbox");
			tb.setAttribute("style","width:7ex");
			tb.setAttribute("disabled","true");
			if (cols[i][info[1]["notnull"][0]] != 0)
				tb.setAttribute("checked", true);
			hbox.appendChild(tb);

			var tb = document.createElement("textbox");
			tb.setAttribute("style","width:10ex");
			tb.setAttribute("readonly","true");
			var def_val = cols[i][info[1]["dflt_value"][0]];
			//if (info[2][i][info[1]["dflt_value"][0]] == 3)
			//	def_val = sm_quote(def_val);
			tb.setAttribute("value", def_val);
			hbox.appendChild(tb);

			var tb = document.createElement("button");
			tb.setAttribute("label", sm_getLStr("dropColumn"));
			tb.setAttribute("value", cols[i][info[1]["name"][0]]);
			tb.setAttribute("oncommand", "SQLiteManager.modifyTable('dropColumn','" + sTable + "','" + cols[i][info[1]["name"][0]] + "')");
			hbox.appendChild(tb);

			vbox.appendChild(hbox);
		}
		
		var aTableInfo = Database.getTableInfo(sTable, -1);

		document.getElementById("numFields").value = aTableInfo["numFields"];
		document.getElementById("numIndexes").value = aTableInfo["numIndexes"];
		document.getElementById("numRecords").value = aTableInfo["numRecords"];
  },

  printIndexInfo: function(sIndex)
  {
		var aIndexInfo = Database.getIndexInfo(sIndex);
		
		document.getElementById("tabletoindex").value = aIndexInfo["table"];
		document.getElementById("fieldtoindex").value = aIndexInfo["cols"];
		document.getElementById("duplicatevalues").value = sm_getLStr("allowed");
		if(aIndexInfo["unique"] == 1)
			document.getElementById("duplicatevalues").value = sm_getLStr("notAllowed");
  },

	//loadTabBrowse: populates the table list and the tree view for current table
	//must be called whenever a database is opened/closed
	//and whenever the schema changes
	//depends entirely upon the values in "browse-type" and "browse-name" controls
  loadTabBrowse: function()
  {
    //no need to waste resources if this tab is not selected
    if(this.getSelectedTabId() != "tab-browse")
    	return false;
    	
    if (this.sCurrentDatabase == null)
    	return false;

		if (this.mostCurrObjType == null)
			return false;

		var sObjType = this.mostCurrObjType.toLowerCase();
		if (sObjType != "table" && sObjType != "view")
			return false;

		document.getElementById("browse-type").value = sObjType.toUpperCase();
		document.getElementById("browse-name").value = this.mostCurrObjName;

		//populate the treeview
    var sObjName = this.mostCurrObjName;
		if (sObjName != this.sBrowseObjName)
		{
			this.miOffset = 0;
			this.sBrowseObjName = sObjName;
			this.msBrowseCondition = "";
		}
			
		//some functionality in the UI depends on whether table/view is shown
		var btnAdd =	document.getElementById("btnAddRecord");
		var btnEdit =	document.getElementById("btnEditRecord");
		var btnDelete =	document.getElementById("btnDeleteRecord");
		//two search buttons because set attribute cannot change oncommand attribute
		var btnSearchTable =	document.getElementById("btnSearchTable");
		var btnSearchView =	document.getElementById("btnSearchView");
		var treeChildren = document.getElementById("browse-treechildren");
		btnSearchTable.hidden = true;
		btnSearchView.hidden = true;
		if (sObjType == "table")
		{
			btnAdd.disabled = false;
			btnEdit.disabled = false;
			btnDelete.disabled = false;
			treeChildren.setAttribute("context", "menupopup-editTableRow");
			btnSearchTable.hidden = false;
		}
		if (sObjType == "view")
		{
			btnAdd.disabled = true;
			btnEdit.disabled = true;
			btnDelete.disabled = true;
			treeChildren.setAttribute("context", "");
			btnSearchView.hidden = false;
		}

    TreeTable.Initialize("browse-tree");
    TreeTable.ShowTable(false);

    try {
      var bRet = Database.loadTableData(sObjType, sObjName, this.msBrowseCondition,   
										this.miLimit, this.miOffset);
    } catch (e) 
    { 
      sm_message(sm_getLStr("loadDataFailed"), 0x3);
			return false;
    }
    if (!bRet)
    	return bRet;

    var records = Database.getRecords();
    var columns = Database.getColumns();
  	this.miCount = Database.getTableCount(sObjName, this.msBrowseCondition);
    	
		this.manageNavigationControls();    	
    if (records && columns)
        TreeTable.PopulateTableData(records, columns);
		return true;
  },

	onBrowseNavigate: function(sType)
	{
		switch(sType)
		{
			case "first":
				this.miOffset = 0;
				break;
			case "previous":
				this.miOffset = this.miOffset - this.miLimit;
				if (this.miOffset < 0)
					this.miOffset = 0;
				break;
			case "next":
				this.miOffset = this.miOffset + this.miLimit;
				break;
			case "last":
				this.miOffset = this.miCount - (this.miCount % this.miLimit);
				break;
		}
		this.loadTabBrowse();
	},
	
	manageNavigationControls: function()
	{
		//manage textboxes
  	document.getElementById("nav-total-val").value = this.miCount;
  	var iStart = (this.miCount == 0) ? 0 : (this.miOffset + 1);
  	document.getElementById("nav-start-val").value = iStart;
  	var iEnd = this.miOffset + this.miLimit;
  	iEnd = ((iEnd > this.miCount) || (this.miLimit == -1)) ? this.miCount : iEnd;
  	document.getElementById("nav-end-val").value = iEnd;

		//manage buttons
		var btnFirst = document.getElementById("btn-nav-first");
		var btnPrevious = document.getElementById("btn-nav-previous");
		var btnNext = document.getElementById("btn-nav-next");
		var btnLast = document.getElementById("btn-nav-last");

		btnFirst.disabled = false;
		btnPrevious.disabled = false;
		btnNext.disabled = false;
		btnLast.disabled = false;

    //manage the navigate buttons
    if (this.miLimit < 0 || this.miLimit >= this.miCount)
    {
			btnFirst.disabled = true;
			btnPrevious.disabled = true;
			btnNext.disabled = true;
			btnLast.disabled = true;
			return;
    }

		if (this.miOffset == 0)
		{
			btnFirst.disabled = true;
			btnPrevious.disabled = true;
		}
		else
		{
			btnFirst.disabled = false;
			btnPrevious.disabled = false;
		}
		if (this.miOffset + this.miLimit > this.miCount)
		{
			btnNext.disabled = true;
			btnLast.disabled = true;
		}
		else
		{
			btnNext.disabled = false;
			btnLast.disabled = false;
		}
	},
	
	//loadTabExecute: anything to be done when that tab is shown goes here
  loadTabExecute: function()
  {
	  //alert("dbinfo");
  },

	//loadTabDbInfo: anything to be done when that tab is shown goes here
  loadTabDbInfo: function()
  {
    //no need to waste resources if this tab is not selected
    if(this.getSelectedTabId() != "tab-dbinfo")
    	return false;
    	
    if (this.sCurrentDatabase == null)
    	return false;

		var aSettings = ["auto_vacuum", "cache_size", "count_changes", "default_cache_size", "empty_result_callbacks", "encoding", "full_column_names", "fullfsync", "legacy_file_format", "locking_mode", "page_size", "max_page_count", "read_uncommitted", "short_column_names", "synchronous", "temp_store", "temp_store_directory"];
    if (gbGecko_1_9_b3)
		  aSettings = ["schema_version", "user_version", "auto_vacuum", "cache_size", "count_changes", "default_cache_size", "empty_result_callbacks", "encoding", "full_column_names", "fullfsync", "legacy_file_format", "locking_mode", "page_size", "max_page_count", "read_uncommitted", "short_column_names", "synchronous", "temp_store", "temp_store_directory"];
		for(var i = 0; i < aSettings.length; i++)
		{
		  var sSetting = aSettings[i];
			var node = document.getElementById("pr-" + sSetting);
			var newVal = Database.getSetting(sSetting);
			node.value = newVal;
		}
		return true;
	},

  searchView: function()
  {
  	var sViewName = this.aCurrObjNames["view"];
  	Database.loadTableData("view", sViewName, "", 1, 0);
    var records = Database.getRecords();
    if (records.length == 0)
    {
    	alert(sm_getLStr("noRecord"));
    	return false;
    }

		var columns = Database.getColumns();
  	var names = [], types = [];
    for (var col in columns) {
    	names[col] = columns[col][0];
    	types[col] = GetColumnTypeString(columns[col][1]);
    }
    var cols = [names, types];
    window.openDialog("chrome://sqlitemanager/content/tableOperations.xul",
					"tableOperations", "chrome, resizable, centerscreen, modal, dialog", 
					Database, this.aCurrObjNames["view"], "search-view", cols);
    return true;
  },

  //getSelectedTabId: returns the id of the selected tab
  getSelectedTabId: function()
  {
		var oSelectedTab = document.getElementById("sm-tabs").selectedItem;
		var id = oSelectedTab.getAttribute("id");
		return id;
  },

  //doSelectTab: called when onselect event fires on tabs[id="sm-tabs"]
	doSelectTab: function(oSelectedTab)
	{
		var id = oSelectedTab.getAttribute("id");
		return this.loadTabWithId(id);
	},

	loadTabWithId: function(sId)
	{
		switch(sId)
		{
			case "tab-structure":
				this.loadTabStructure();
				break;
			case "tab-browse":
				this.loadTabBrowse();
				break;
			case "tab-execute":
				this.loadTabExecute();
				break;
			case "tab-dbinfo":
				this.loadTabDbInfo();
				break;
		}
		return true;
	},

  useExtensionManagementTable: function(bUse, bImplicit)
  {
    if (typeof bUse == "string")
      bUse = (bUse=="true")?true:false;

		if(this.sCurrentDatabase == null)
		{
			alert(sm_getLStr("firstOpenADb"));
			return false;
		}

    smExtManager.setUsage(bUse, bImplicit);

		//refresh the structure tree here so that mgmt table is shown/removed
		this.refresh();

		//toggle the label of the menuitem
    var mi = document.getElementById("menu-general-extensionTable");
    var lblEnable = sm_getLStr("enableExtDataTable");
    var lblDisable = sm_getLStr("disableExtDataTable");
    mi.setAttribute("label", (bUse?lblDisable:lblEnable));
//    mi.setAttribute("oncommand", str);//does not work
    mi.setAttribute("value", !bUse);

		//hide/show the images for query history in the execute sql tab
		document.getElementById("queryHistoryPrevImage").hidden = !bUse;
		document.getElementById("queryHistoryNextImage").hidden = !bUse;
		document.getElementById("queryHistoryClearImage").hidden = !bUse;

    return true;
  },

  showPrevSql: function()
  {
    var sQuery = smExtManager.getPrevSql();
    if (!sQuery) return;
    document.getElementById("txtSqlStatement").value = sQuery;
  },

  showNextSql: function()
  {
    var sQuery = smExtManager.getNextSql();
    if (!sQuery) return;
    document.getElementById("txtSqlStatement").value = sQuery;
  },

  clearSqlHistory: function()
  {
    smExtManager.clearSqlHistory();
  },

	runSqlStatement: function(sType)
	{
		if(this.sCurrentDatabase == null)
		{
			alert(sm_getLStr("firstOpenADb"));
			return;
		}

		//get the query string from an xul page
		var sQuery = document.getElementById("txtSqlStatement").value;
		
	  var re_ws = /\t*\r*\n*\s*/
		var a = re_ws.exec(sQuery);
		if(sQuery.length == a[0].length)
		{
			alert(sm_getLStr("writeSomeSql"));
			return;
		}
		//alert(sQuery);
		if(sType == "select")
		{
			var bRet = Database.selectQuery(sQuery);
			//Database.executeMultiple(sQuery);
      //store the query in config db
			if (bRet) smExtManager.addQuery(sQuery);
      //set this value so that query history is reset to latest query
      //that is previous will again begin from the latest query
  		smExtManager.goToLastQuery();
		}
		if(sType == "execute")
		{
			Database.executeTransaction([sQuery]);
		}
		var aData = Database.getRecords();
		var aColumns = Database.getColumns();
		//display the last error in the textbox
		document.getElementById("sqlLastError").value = Database.getLastError();
		sm_message(sm_getLStr("rowsReturned") + ": " + aData.length, 0x2);
		//document.getElementById("sqlRowsAffected").value = " ( " + aData.length + " rows returned )";
		
		//the following two lines must be before the code for tree
		//otherwise, it does not refresh the structure tree as expected
		this.loadDBStructure();
		this.loadTabBrowse();
		
		TreeTable.Initialize("treeSqlOutput");
		TreeTable.ShowTable(false);
		TreeTable.Initialize("treeSqlOutput");
		TreeTable.PopulateTableData(aData, aColumns);
	},
		
  newDatabase: function()
  {
  	var sExt = "." + sm_getPreferenceValue("sqliteFileExtensions", "sqlite");
		//prompt for a file name
		var fname = prompt("Enter the database name (" + sExt + " will be automatically appended to the name)", "", "Enter the Database Name");

		//if cancelled, abort
		if (fname == "" || fname == null)
			return false;
		
		//append the extension to the chosen name
		fname += sExt;
		
		//let the user choose the folder for the new db file	
		var dir = sm_chooseDirectory(sm_getLStr("selectFolderForDb"));
		if (dir != null) 
		{
			//access this new copied file
			var newfile = Components.classes["@mozilla.org/file/local;1"]
								.createInstance(Components.interfaces.nsILocalFile);
			newfile.initWithPath(dir.path);
			newfile.append(fname);
			
			//if the file already exists, alert user that existing file will be opened 
			if(newfile.exists())
			{
				alert(sm_getLStr("dbFileExists"));
			}

    	//if another file is already open,
			//confirm from user that it should be closed
    	if(this.closeDatabase(false))
    	{
				//assign the new file (nsIFile) to the current database
				this.sCurrentDatabase = newfile;
				//if the file does not exist, openDatabase will create it 
				this.setDatabase(this.sCurrentDatabase);
				return true;
			}
		}
		return false;
  },

	//closeDatabase: 
  closeDatabase: function(bAlert)
  {
		//nothing to close if no database is already open    
  	if(this.sCurrentDatabase == null)
  	{
  		if(bAlert)
				alert(sm_getLStr("noOpenDb"));
			return true;
		}
			
   	//if another file is already open, confirm before closing
		var ans = confirm(sm_getLStr("confirmClose"));
		if(!ans)
			return false;
					
		//make the current database as null and 
		//call setDatabase to do appropriate things
		this.sCurrentDatabase = null;
		this.setDatabase(this.sCurrentDatabase);
		return true;
	},
		
  copyDatabase: function()
  {
    var sExt = "." + sm_getPreferenceValue("sqliteFileExtensions", "sqlite");
 		if(this.sCurrentDatabase == null)
 		{
			alert(sm_getLStr("firstOpenADb"));
			return;
		}
		//prompt for a file name
		var fname = prompt("Enter the database name (." + sExt + " will be automatically appended to the name)", "", "Enter the Database Name");

		//if cancelled, abort
		if (fname == "" || fname == null)
			return;
		else
			fname += sExt;
			
		//let the user choose the folder for the new db file	
		//let the user choose the folder for the new db file	
		var dir = sm_chooseDirectory(sm_getLStr("selectFolderForDb"));
		if (dir != null) 
		{
			//copy the opened file to chosen location
			this.sCurrentDatabase.copyTo(dir, fname);

			//access this new copied file
			var newfile = Components.classes["@mozilla.org/file/local;1"]
								.createInstance(Components.interfaces.nsILocalFile);
			newfile.initWithPath(dir.path);
			newfile.append(fname);
			
			//if the file does not exist, openDatabase will create it 
			if(!newfile.exists())
			{
				var ans = confirm(sm_getLStr("copyFailed"));
				if(!ans)
					return;
			}

			//assign the new file (nsIFile) to the current database
    	if(this.closeDatabase(false))
    	{
				this.sCurrentDatabase = newfile;
        this.setDatabase(this.sCurrentDatabase);
        return;
			}
		}
		return;
  },
    
  compactDatabase: function()
  {
    if (this.sCurrentDatabase == null)
		{
			alert(sm_getLStr("firstOpenADb"));
			return false;
		}
  	var sQuery = "VACUUM";
  	//cannot vacuum from within a transaction
  	Database.selectQuery(sQuery);
  	return true;
  },

  analyzeDatabase: function()
  {
    if (this.sCurrentDatabase == null)
    {
			alert(sm_getLStr("firstOpenADb"));
			return false;
    }
  	var sQuery = "ANALYZE";
  	//cannot vacuum from within a transaction
  	Database.selectQuery(sQuery);
  	return true;
  },

  checkIntegrity: function()
  {
    if (this.sCurrentDatabase == null)
    {
			alert(sm_getLStr("firstOpenADb"));
			return false;
    }
    Database.selectQuery("PRAGMA integrity_check");
    var records = Database.getRecords();
    var columns = Database.getColumns();

    var txt = sm_getLStr("integrityResultPrefix") + ": ";
    //report OK if i row returned containing the value "ok"
    if (records.length == 1 && records[0][0] == "ok")
      alert(txt + sm_getLStr("ok"));
    else
      alert(txt + sm_getLStr("notOk"));
    return true;
  },

  openDatabase: function()
  {        
    var sExt = "." + sm_getPreferenceValue("sqliteFileExtensions", "sqlite");

		const nsIFilePicker = Components.interfaces.nsIFilePicker;
		var fp = Components.classes["@mozilla.org/filepicker;1"]
			           .createInstance(nsIFilePicker);
		fp.init(window, sm_getLStr("selectDb"), nsIFilePicker.modeOpen);
		fp.appendFilter(sm_getLStr("sqliteDbFiles") + " (*" + sExt + ")", "*" + sExt + "");
		fp.appendFilters(nsIFilePicker.filterAll);
		
		var rv = fp.show();
		if (rv == nsIFilePicker.returnOK || rv == nsIFilePicker.returnReplace) 
		{
		  // work with returned nsILocalFile...
    	if(this.closeDatabase(false))
    	{
				this.sCurrentDatabase = fp.file;
        this.setDatabase(this.sCurrentDatabase);
        return true;
			}
		}
		return false;
  },

  saveDatabase: function()
  {        
  },

  createTable: function()
  {        
    if (this.sCurrentDatabase != null)
		{
      window.openDialog("chrome://sqlitemanager/content/createTable.xul",
							"createTable", "chrome, resizable, centerscreen, modal, dialog", 
							Database, null);
			this.loadDBStructure();
			this.loadTabBrowse();
		}
		else
		{
			alert(sm_getLStr("firstOpenADb"));
		}
  },

  modifyTable: function(sOperation, sTable, sColumn)
  {        
  	if(sOperation == "dropColumn")
  	{
			var bReturn = StructureManager.modifyTable("dropColumn", sTable, sColumn);
			if(bReturn)
			{
				this.loadDBStructure();
				this.loadTabBrowse();
			}
			return bReturn;
		}
		return false;
  },

  reindexIndex: function()
  {        
  	var sCurrIndex = this.aCurrObjNames["index"];
  	if(sCurrIndex != null && sCurrIndex != undefined && sCurrIndex.length > 0)
  	{
			var bReturn = StructureManager.reindexObject("INDEX", sCurrIndex);
			return bReturn;
		}
		return false;
  },

	dropObject: function(sObjectType)
  {
    if (this.sCurrentDatabase == null)
		{
			alert(sm_getLStr("firstOpenADb"));
			return false;
		}

		var sObjectName = "";
		sObjectName = this.aCurrObjNames[sObjectType];
		
		var aNames = this.aObjNames[sObjectType];

		if(aNames.length == 0)
		{
			alert(sm_getLStr("noObjectToDelete") + ": " + sObjectType);
			return false;
		}
		var bReturn = StructureManager.dropObject(sObjectType, sObjectName);
		if(bReturn)
		{
			sm_message(sm_getLStr("dropDone"), 0x2);
			this.loadDBStructure();
			this.loadTabBrowse();
		}
		return bReturn;
  },

	importFromFile: function()
	{
    if (this.sCurrentDatabase == null)
		{
			alert(sm_getLStr("firstOpenADb"));
			return false;
		}
    window.openDialog("chrome://sqlitemanager/content/export.xul",
						"import",	"chrome, resizable, centerscreen, modal, dialog", 
						Database, "import");
		this.loadDBStructure();
		this.loadTabBrowse();
		return true;
	},
	
	exportObject: function(sObjectType)
  {
    if (this.sCurrentDatabase == null)
		{
			alert(sm_getLStr("firstOpenADb"));
			return false;
		}

		var sObjectName = "";
		sObjectName = this.aCurrObjNames[sObjectType];
		
		var aNames = this.aObjNames[sObjectType];

		if(aNames.length == 0)
		{
			alert("No " + sObjectType + " to export");
			return false;
		}
    window.openDialog("chrome://sqlitemanager/content/export.xul",
						"export",	"chrome, resizable, centerscreen, modal, dialog", 
						Database, sObjectType, sObjectName);
		return true;
  },

  createObject: function(sObjectType)
  {
    if (this.sCurrentDatabase == null)
		{
			alert(sm_getLStr("firstOpenADb"));
			return false;
		}

		var xul = "chrome://sqlitemanager/content/create" + sObjectType + ".xul";
    var ret = window.openDialog(xul, "create" + sObjectType, 
						"chrome, resizable, centerscreen, modal, dialog", 
						Database, this.aCurrObjNames["table"], sObjectType);
		this.loadDBStructure();
		this.loadTabBrowse();
		return true;
  },

// operateOnTable: various operations on a given table
// sOperation = rename | copy | reindex | delete	| 
//              insert | search | update
  operateOnTable: function(sOperation)
  {
		//these operations make sense in the context of some table
		//so, take action only if there is a valid selcted db and table
	  if (this.sCurrentDatabase == null || this.aCurrObjNames["table"] == null)
		{
			alert(sm_getLStr("noDbOrTable"));
			return false;
		}
  	var sCurrTable = this.aCurrObjNames["table"];
  	var bReturn = false;
  	var bRefresh = false; //to reload tabs
		switch(sOperation)
		{
			case "reindex":
				return StructureManager.reindexObject("TABLE", sCurrTable);
				break;
			case "analyze":
			  return StructureManager.analyzeTable(sCurrTable);
				break;
		}
  	if(sOperation == "copy")
  	{
			var bReturn = StructureManager.copyTable(sCurrTable);
			if(bReturn)
			{
				this.loadDBStructure();
				this.loadTabBrowse();
			}
			return bReturn;
		}
  	if(sOperation == "rename")
  	{
			var bReturn = StructureManager.renameTable(sCurrTable);
			if(bReturn)
			{
				this.loadDBStructure();
				this.loadTabBrowse();
			}
			return bReturn;
		}
  	if(sOperation == "drop")
  	{
			var bReturn = StructureManager.dropObject("TABLE", sCurrTable);
			if(bReturn)
			{
				this.loadDBStructure();
				this.loadTabBrowse();
			}
			return bReturn;
		}
  	if(sOperation == "empty")
  	{
			var bReturn = StructureManager.emptyTable(sCurrTable);
			if(bReturn)
			{
				this.loadDBStructure();
				this.loadTabBrowse();
			}
			return bReturn;
		}
  	if(sOperation == "addColumn")
  	{
  		var newCol = [];
			newCol["name"] = document.getElementById("tb-addcol-name").value;
			newCol["type"] = document.getElementById("tb-addcol-type").value;
			newCol["notnull"] = document.getElementById("tb-addcol-notnull").checked;
			newCol["dflt_value"] = document.getElementById("tb-addcol-default").value;
			newCol["dflt_value"] = sm_makeSqlValue(newCol["dflt_value"]);
			var bReturn = StructureManager.modifyTable("addColumn", sCurrTable, newCol);
			if(bReturn)
			{
				document.getElementById("tb-addcol-name").value = "";
				document.getElementById("tb-addcol-type").value = "";
				document.getElementById("tb-addcol-notnull").checked = false;
				document.getElementById("tb-addcol-default").value = "";
				this.loadDBStructure();
				this.loadTabBrowse();
			}
			return bReturn;
		}
			
		//update the first selected row in the tree, else alert to select
  	//if selection exists, pass the rowid as the last arg of openDialog
  	var aRowIds = [];
  	var rowCriteria = "";
  	if(sOperation == "update" || sOperation == "delete")
  	{
  		var colMain = Database.getTableRowidCol(this.aCurrObjNames["table"]);
			colMain["name"] = "`" + colMain["name"] + "`";

  		//allowing for multiple selection in the tree
			var tree = document.getElementById("browse-tree");
			var start = new Object();
			var end = new Object();
			var numRanges = tree.view.selection.getRangeCount();

			for (var t = 0; t < numRanges; t++){
				tree.view.selection.getRangeAt(t,start,end);
			  for (var v = start.value; v <= end.value; v++){
			  	var rowid = tree.view.getCellText(v,
							tree.columns.getColumnAt(colMain["cid"]));
					aRowIds.push(rowid);
			  }
			}
			//do nothing, if nothing is selected
			if(aRowIds.length == 0)
			{
				alert(sm_getLStr("noRecord"));
				return false;
			}
			//if editing, should select only one record
			if (sOperation == "update")
			{
				if (aRowIds.length != 1)
				{
					alert(sm_getLStr("onlyOneRecord"));
					return false;
				}
				rowCriteria = " " + colMain["name"] + " = " + aRowIds[0];
			}
			//if deleting, pass as argument rowid of all selected records to delete
			if (sOperation == "delete")
			{
				rowCriteria = " " + colMain["name"] + " = " + rowid;
				var bReturn = StructureManager.deleteFromTable(sCurrTable,
								colMain["name"], aRowIds);
				if(bReturn)
					this.loadTabBrowse();
				return bReturn;
			}
		}

    window.openDialog("chrome://sqlitemanager/content/tableOperations.xul",
						"tableOperations", 
						"chrome, resizable, centerscreen, modal, dialog", 
						Database, this.aCurrObjNames["table"], sOperation, rowCriteria);
		if(sOperation != "search")
		{
			if(sOperation != "update")
			{
				this.loadDBStructure();
			}
			this.loadTabBrowse();
		}
		return true;
  },

  // populateProfileDBList: Load list of files with extension passed as .arg
  populateProfileDBList: function(extSqlite)
  {
    // Get the nsIFile object pointing to the profile directory
    var fileList = Components.classes["@mozilla.org/file/directory_service;1"]
            .getService(Components.interfaces.nsIProperties)
            .get("ProfD", Components.interfaces.nsIFile)
            .directoryEntries; 
    var file;
    var iFileCount = 0;
    var iPosition;
    
    const extLength = extSqlite.length;
    
    var aItems = new Array();
    aItems.push(sm_getLStr("selectProfileDb"));

		while (fileList.hasMoreElements()) 
    {
      file = fileList.getNext().QueryInterface(Components.interfaces.nsIFile);
      iPosition = file.leafName.lastIndexOf(extSqlite);
      if (iPosition > 0 && iPosition == file.leafName.length - extLength)
      {
        iFileCount = iFileCount + 1;
        aItems.push(file.leafName);
      } 
    }
    //get the node for the popup menus to show profile db list
    var listboxProfileDB = document.getElementById("listbox-profileDB");

    PopulateDropDownItems(aItems, listboxProfileDB, null);
    sm_message(sm_getLStr("filesInProfileDbList") + ": " + iFileCount, 0x2);
  },

  // openProfileDatabase: open a file from the database dropdown list
  openProfileDatabase: function()
  { 
		//get the node for dropdown menu in which profile db list is shown       
  	var listboxProfileDB = document.getElementById("listbox-profileDB");
    var sLabel = listboxProfileDB.value;

    var file = Components.classes["@mozilla.org/file/directory_service;1"]
            .getService(Components.interfaces.nsIProperties)
            .get("ProfD", Components.interfaces.nsIFile);
    file.append(sLabel);
		
		//proceed only if the file exists
		//we are in the profile folder via the listbox, so open if the file exists
		//do not attempt to create new file
		if(!file.exists())
		{
			alert(sm_getLStr("invalidProfileDb"));
			return false;
		}
		
  	if(this.closeDatabase(false))
  	{
			this.sCurrentDatabase = file;
      this.setDatabase(this.sCurrentDatabase);
      return true;
		}
		return false;
	},

  // setDatabase: set the current database to nsiFileObj
  setDatabase: function(nsiFileObj) 
  {
  //when passed as arg, works but fails to show .path and .leafName properties
//      this.sCurrentDatabase = nsiFileObj; 

		TreeTable.Initialize("browse-tree");
		TreeTable.ShowTable(false);
		TreeTable.Initialize("treeSqlOutput");
		TreeTable.ShowTable(false);
		
		//try connecting to database
		var bConnected = false;
		try
		{
			if(this.sCurrentDatabase != null)
		    bConnected = Database.ConnectToDatabase(this.sCurrentDatabase);
			if(this.sCurrentDatabase == null)
		    Database.closeConnection();
		} 
		catch (e)
		{
	    var sTemp = this.sCurrentDatabase.path;
	    this.sCurrentDatabase = null;
	    sm_message("Connect to '" + sTemp + "' failed: " + e, 0x3);
	    return;
		}

    var path = "", leafName = "";
    if (nsiFileObj != null && bConnected)
    {
      path = nsiFileObj.path;
      leafName = nsiFileObj.leafName;
      //change the mruPath.1 in preferences
      sm_setPreferenceValue("mruPath.1", path);

      //extension related mgmt info
      if (this.m_bMgmtDataIsInDb)
      {
        smExtManager = new SMExtensionManager();
  //  else
  //    smExtManager = new SMExtensionManagerExternal();
  
        var bUsage = smExtManager.getUsage();
        this.useExtensionManagementTable(bUsage, true);
      }
    }
    if (!bConnected)
    {
    	this.sCurrentDatabase = null;
    }
    //change window title to show db file path
		document.title = sm_getLStr("extName") + " - " + path;

		document.getElementById("col-table-view").setAttribute("label",leafName);
		//reload the two tabs
    this.loadDBStructure();
  },
    
  onSqliteVersion: function()
  {
    //connection is required for fetching the SQLite library version too
    if (this.sCurrentDatabase == null)
    {
      alert(sm_getLStr("firstOpenADb"));
      return false;
    }
    alert(sm_getLStr("sqliteVersion") + ": " + Database.getSqliteVersion());
    return true;
  },
	
  setSqlText: function(val)
  {
    document.getElementById("txtSqlStatement").value = val;
  }
};

// ****** table event handling and display ******
var TreeTable = 
{    
    treeTable: null,    // Tree containing listing of current table
    lastRow: null,      // Last row that was clicked on (may be inaccurate if scrolling happened)
    lastCol: null,      // Last col that was clicked on (may be inaccurate if scrolling happened)

    // Initialize: Set up the treeview which will display the table contents
    Initialize: function(sTableID) 
    {        
        this.treeTable = document.getElementById(sTableID);
    },

    // ShowTable: Show/hide any currently displayed table data
    ShowTable: function(bStatus) 
    {
        if (this.treeTable == null) 
            return;
        if (bStatus) // Show the tree
        {
            this.treeTable.childNodes[0].style.display = "";
            this.treeTable.childNodes[1].style.display = "";
        } else // Hide the tree and remove all of the child rows/columns
        {
            ClearElement(this.treeTable.childNodes[0]);
            ClearElement(this.treeTable.childNodes[1]);
            this.treeTable.childNodes[0].style.display = "none";
            this.treeTable.childNodes[1].style.display = "none";
        } 
    },
    
    // UserCopyCell: Handle user choosing to copy the cell contents
    UserCopyCell: function()
    {
        var sCell;
        if (this.lastRow && this.lastCol)
        {
            sCell= this.treeTable.view.getCellText(this.lastRow, this.lastCol);
            if (sCell)
                SQLiteManager.clipService.copyString(sCell);
        }
    },

    // UserCopyRows: Handle user choosing to copy all currently highlighted rows
    UserCopyRows: function()
    {
        var sel = this.treeTable.view.selection;
        var start = new Object();
        var end = new Object();
        var iRangeCount = sel.getRangeCount();
        var sData = "";
        for (var iRange = 0; iRange < iRangeCount; iRange++)
				{
          sel.getRangeAt(iRange,start,end);
          for (var iRow = start.value; iRow <= end.value; iRow++)
					{            
            if (sData.length > 0)
              sData = sData + "\n";
            sData = sData + this.GetRowData(iRow, "csv");
          }
        }
        if (sData)
            SQLiteManager.clipService.copyString(sData);
    },

    // GetRowData: Get the contents of an entire numbered row, using the specified format
    GetRowData: function(iRow, sFormat) 
    {
        var result= null;
        if (sFormat=="csv")
            result="";

        return result;       
    },

    // UserTreeClick: Handle the user clicking on the tree
    UserTreeClick: function(ev)
    {
        // This event happens AFTER the window has scrolled (if necessary). This means that
        // if the user clicked on an element that is partially off screen, and the screen scrolls
        // to fully display it, then the mouse may no longer be over the correct element.
        //if (ev.button==2) // Right-click
            this.StoreClickLocation(ev);
    },

  // AddTreecol: Add the column to the tree view
  AddTreecol: function(treecols, sId, col, sTooltip, iWidth, bLast) 
  {
    var treecol = document.createElement("treecol");
    treecol.setAttribute("label", col);
    treecol.setAttribute("id", sId);
    if (sTooltip != null)
      treecol.setAttribute("tooltiptext", sTooltip);

    if (iWidth > 0)
      treecol.setAttribute("style", "width:" + iWidth + "ex");
		treecol.setAttribute("flex","1");
    if (bLast)
    {
			//treecol.removeAttribute("flex");
			//this minwidth setting does not seem to work.
			//also, there is no way yet of resizing the last column
			treecol.setAttribute("minwidth",iWidth);
    }
    treecols.appendChild(treecol); 

    //if (!bLast) 
		{
      var splitter = document.createElement("splitter");
      splitter.setAttribute("id", "splitter" + col);
      splitter.setAttribute("class", "tree-splitter");        
      splitter.setAttribute("resizebefore", "closest");
      splitter.setAttribute("resizeafter", "closest");
      treecols.appendChild(splitter); 
    }
  },
  AddTreecol3: function(treecols, sId, col, sTooltip, iWidth, bLast) 
  {
    var treecol = document.createElement("treecol");
    treecol.setAttribute("label", col);
    treecol.setAttribute("id", sId);
		treecol.setAttribute("width",iWidth);
		treecol.setAttribute("minwidth",10);
    if (sTooltip != null)
      treecol.setAttribute("tooltiptext", sTooltip);

    if (bLast)
    {
      //want to do anything special for the last column? do it here.
    }
    treecols.appendChild(treecol);

    //add a splitter after every column
    var splitter = document.createElement("splitter");
    splitter.setAttribute("id", "splitter" + col);
    splitter.setAttribute("class", "tree-splitter");        
    splitter.setAttribute("resizebefore", "closest");
    splitter.setAttribute("resizeafter", "grow");
    treecols.appendChild(splitter); 
  },

  // StoreClickLocation: Store the row/column that the click occurred; used later when copying cell text
  StoreClickLocation: function(ev)
  {
    // http://www.xulplanet.com/references/elemref/ref_tree.html
    if (ev && this.treeTable && ev.type == "click")
    {
      var row = {}, col = {}, obj = {};
      this.treeTable.treeBoxObject.getCellAt(ev.clientX, ev.clientY, row, col, obj);            
      if (row && row.value && col && col.value!=null) // Clicked on a cell
      {
        this.lastRow= row.value;
        this.lastCol= col.value;
      } else
      {
        this.lastRow= null;
        this.lastCol= null;
      }
    }
  },

  // PopulateTableData: Assign our custom treeview
  PopulateTableData: function(aTableData, aColumns)
  {   
      var treecols = this.treeTable.firstChild;
      ClearElement(treecols);
      
      // Order columns (slow method; needs to be reworked)
      var iColumnCount = 0;
      var iRow;
      var iWidth, iTotalWidth, iMaxWidth;
      var sTooltip;
      var allCols = [];
      for (var col in aColumns) 
      {
        iColumnCount = iColumnCount + 1;
        var aTemp = [aColumns[col][0], aColumns[col][1]];
        allCols.push(aTemp);
			}

      var iTreeWidth = this.treeTable.boxObject.width;
      for (var iColumn = 0; iColumn < iColumnCount; iColumn++)
			{
        iTotalWidth = 0;
        iMaxWidth = 0;
        iTotalWidth = iTreeWidth/iColumnCount;
        if (iTotalWidth < 50) iTotalWidth = 50;

        sTooltip = allCols[iColumn][0] + " (" + GetColumnTypeString(allCols[iColumn][1]) + ")";
        
        if (gbGecko_1_9_b3)
          this.AddTreecol3(treecols, iColumn, allCols[iColumn][0], sTooltip, iTotalWidth, (iColumn==iColumnCount-1?true:false));
        else
          this.AddTreecol(treecols, iColumn, allCols[iColumn][0], sTooltip, iTotalWidth, (iColumn==iColumnCount-1?true:false));
      }
      this.treeTable.view = new this.DatabaseTreeView(aTableData,aColumns);
      this.ShowTable(true);
  },

    // DatabaseTreeView: Create a custom nsITreeView - good for displaying large amounts of data
    DatabaseTreeView: function(aTableData, aColumns) 
    {
        // http://kb.mozillazine.org/Sorting_Trees
        // 2 dimensional array containing table contents
        this.aTableData = aTableData; 
        // Column information (index, order, type)
				this.aColumns = aColumns;   
				// Number of rows in the table
        this.rowCount = aTableData.length;  

        this.getCellText = function(row,column) 
        { 
            var sResult;
            try 
						{ sResult= this.aTableData[row][column.id]; }
            catch (e) 
						{ return "<" + row + "," + column.id + ">"; }
            return sResult;
        };

        this.setTree             = function(treebox){ this.treebox=treebox; };
        this.isContainer         = function(row){ return false; };
        this.isSeparator         = function(row){ return false; };
        this.isSorted            = function(row){ return false; };
        this.cycleHeader         = function(col){ this.SortColumn(col); }
        this.getLevel            = function(row){ return 0; };
        this.getImageSrc         = function(row,col){ return null; };
        this.getRowProperties    = function(row,props){};
        this.getCellProperties   = function(row,col,props){};
        this.getColumnProperties = function(colid,col,props){};

        this.SortColumn          = function(col)
        {
//            var name  = col.id; 
//            var index = this.aColumns[name][0];
//            var order = this.aColumns[name][1];
//            var isnum = ((this.aColumns[name][2]==1 || this.aColumns[name][2]==1)?1:0);
//            this.aColumns[name][1] = (order==0)?1:0; // switch order flag

            var index  = col.id; 
            var name = this.aColumns[index][0];
            var order = this.aColumns[index][1];
            var isnum = ((this.aColumns[index][2]==1 || this.aColumns[index][2]==1)?1:0);
            this.aColumns[index][1] = (order==0)?1:0; // switch order flag

            this.SortTable(this.aTableData,index,order,isnum);  // sort the table
            this.rowCount= this.aTableData.length; // Not the right place for this but...
            //this.treeTable.view = new treeView(this.aTableData, this.aColumns, this.rowCount);       
        };

        // This is the actual sorting method, extending the array.sort() method
        this.SortTable           = function(table,col,order,isnum)
        {
            if(isnum){ // use numeric comparison 
                if(order==0){ // ascending 
                    this.columnSort= function (a,b){ return (a[col]-b[col]); };
                }
                else{ // descending 
                    this.columnSort= function (a,b){ return (b[col]-a[col]); };
                }
            }
            else{ // use string comparison 
                if(order==0){ // ascending 
                    this.columnSort= function(a,b){
                        return (a[col]<b[col])?-1:(a[col]>b[col])?1:0; };
                }
                else{ // descending 
                    this.columnSort= function(a,b){
                        return (a[col]>b[col])?-1:(a[col]<b[col])?1:0; };
                }
            }
            // use array.sort(comparer) method
            table.sort(this.columnSort);
        };
    }
};

//observer for something
function smExampleObserver()
{
  this.register();
}

smExampleObserver.prototype = 
{
  observe: function(subject, topic, data) 
	{
     // Do your stuff here.
  },
  
	register: function() 
	{
    var observerService = Components.classes["@mozilla.org/observer-service;1"]
                          .getService(Components.interfaces.nsIObserverService);
    observerService.addObserver(this, "myTopicID", false);
  },
  
	unregister: function() 
	{
    var observerService = Components.classes["@mozilla.org/observer-service;1"]
                            .getService(Components.interfaces.nsIObserverService);
    observerService.removeObserver(this, "myTopicID");
  }
}

/////////////////////////////////////////////
// following object for db structure tree based on code from 
// http://www.xulplanet.com/tutorials/xultu/treeviewdet.html
var TreeViewDb = {
  removeChildData: function()
  {
    this.visibleData = [[sm_getLStr("noDb"), true, false]];
  },
  
	setChildData: function(aObjects)
	{
	  var aTables = aObjects["table"];
	  var aIndexes = aObjects["index"];
	  var aViews = aObjects["view"];
	  var aTriggers = aObjects["trigger"];

		this.childData = {
			Table: aTables,
			View: aViews,
			Index: aIndexes,
			Trigger: aTriggers
		};
		
		//[celltext, isContainer, isContainerOpen, ...]
		this.visibleData = [
			["Tables (" + aTables.length + ")", true, true, "Table"],
			["Views (" + aViews.length + ")", true, true, "View"],
			["Indexes (" + aIndexes.length + ")", true, true, "Index"],
			["Triggers (" + aTriggers.length + ")", true, true, "Trigger"]
		];
		//reverse order because the index of preceding array members
		//does not change as a result of splice
		for(var idx = 3; idx >= 0; idx--)
		{
      var label = this.visibleData[idx][3];
      var toinsert = this.childData[label];
      for (var i = 0; i < toinsert.length; i++) 
			{
        this.visibleData.splice(idx + i + 1, 0, [toinsert[i], false]);
      }
    }
	},

  treeBox: null,
  selection: null,

  get rowCount()                     { return this.visibleData.length; },
  setTree: function(treeBox)         { this.treeBox = treeBox; },
  getCellText: function(idx, column) { return this.visibleData[idx][0]; },
  isContainer: function(idx)         { return this.visibleData[idx][1]; },
  isContainerOpen: function(idx)     { return this.visibleData[idx][2]; },
  isContainerEmpty: function(idx)    { return false; },
  isSeparator: function(idx)         { return false; },
  isSorted: function()               { return false; },
  isEditable: function(idx, column)  { return false; },

  getParentIndex: function(idx) {
    if (this.isContainer(idx)) return -1;
    for (var t = idx - 1; t >= 0 ; t--) {
      if (this.isContainer(t)) return t;
    }
    return -1;
  },
  getLevel: function(idx) {
    if (this.isContainer(idx)) return 0;
    return 1;
  },
  hasNextSibling: function(idx, after) {
    var thisLevel = this.getLevel(idx);
    for (var t = idx + 1; t < this.visibleData.length; t++) {
      var nextLevel = this.getLevel(t)
      if (nextLevel == thisLevel) return true;
      else if (nextLevel < thisLevel) return false;
    }
    return false;
  },
  toggleOpenState: function(idx) {
    var item = this.visibleData[idx];
    if (!item[1]) return;

    if (item[2]) {
      item[2] = false;

      var thisLevel = this.getLevel(idx);
      var deletecount = 0;
      for (var t = idx + 1; t < this.visibleData.length; t++) {
        if (this.getLevel(t) > thisLevel) deletecount++;
        else break;
      }
      if (deletecount) {
        this.visibleData.splice(idx + 1, deletecount);
        this.treeBox.rowCountChanged(idx + 1, -deletecount);
      }
    }
    else {
      item[2] = true;

      if(this.visibleData[idx][3])
      {
	      var label = this.visibleData[idx][3];
	      var toinsert = this.childData[label];
	      for (var i = 0; i < toinsert.length; i++) {
	        this.visibleData.splice(idx + i + 1, 0, [toinsert[i], false]);
	      }
	      this.treeBox.rowCountChanged(idx + 1, toinsert.length);
	    }
    }
  },

  getImageSrc: function(idx, column) {},
  getProgressMode : function(idx,column) {},
  getCellValue: function(idx, column) {},
  cycleHeader: function(col, elem) {},
  selectionChanged: function() {},
  cycleCell: function(idx, column) {},
  performAction: function(action) {},
  performActionOnCell: function(action, index, column) {},
  getRowProperties: function(idx, column, prop) {},
  getCellProperties: function(idx, column, prop) {},
  getColumnProperties: function(column, element, prop) {}
};