var prefRoot = "extensions.vcssupport.";
var prefManager = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(prefRoot);

var gBundle = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
var mystrings = gBundle.createBundle("chrome://vcssupport/locale/vcssupport.properties");
var vcsFileType = mystrings.GetStringFromName("vcsFileType");

function calVcsExporter() {
	this.wrappedJSObject = this;
}

calVcsExporter.prototype.QueryInterface = function QueryInterface(aIID) {
	if (!aIID.equals(Components.interfaces.nsISupports) && !aIID.equals(Components.interfaces.calIExporter)) {
		throw Components.results.NS_ERROR_NO_INTERFACE;
	}
	return this;
};

calVcsExporter.prototype.getFileTypes = function getFileTypes(aCount) {
	aCount.value = 1;
	var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
	var props = sbs.createBundle("chrome://calendar/locale/calendar.properties");
	return([{defaultExtension:'vcs', extensionFilter:'*.vcs', description:vcsFileType}]);
};

calVcsExporter.prototype.exportToStream = function vcs_exportToStream(aStream, aCount, aItems) {
	//We'ed like to try to open an options dialog here, but
	//Getting an error: Error: openDialog is not defined
	//Source File: --> extensions/%7Befcbec50-6821-11db-bd13-0800200c9a66%7D/components/vcsexport.js : Line: 5
	
	//openDialog("chrome://vcssupport/content/vcssupportOptions.xul", "", "modal")
	
	if(!prefManager.getBoolPref("export_as_ANSI"))	{
		//We can export as UTF-8 for ideographic languages (default behaviour)
		var convStream = Components.classes["@mozilla.org/intl/converter-output-stream;1"].getService(Components.interfaces.nsIConverterOutputStream);
		convStream.init(aStream, 'UTF-8', 0, 0x0000);
	}
	
	var str = "";
	var headers = [];
	// Not using a loop here, since we need to be sure the order here matches
	// with the orders the field data is added later on

	// The structure is STRUCTURE (HEADER) ---  INFO  (EVENTS) --- STRUCTURE (FOOTER)

	headers.push("BEGIN:VCALENDAR");
	headers.push("VERSION:1.0");
	headers.push("PROD-ID:-//Mozilla.org/NONSGML Mozilla Calendar V1.1//EN");
	str = headers.join("\n")+"\n"
	
	if ( prefManager.getBoolPref("export_as_ANSI")) {
		aStream.write(str, str.length);
	} else {
			convStream.writeString(str);
	}
	
	//export_recurrences_as_rules - Decides whether to use RRULE or individual events
	var rulePref = prefManager.getBoolPref("export_recurrences_as_rules");

	for each (item in aItems) {
		var line = [];

		//Use a single function to serialize the events or occurrences
		//Define the variables to be exported in separate functions depending on the event type
		//should be easier to get my head around and save having to repeat code
		if (rulePref) {
			if (item.recurrenceInfo) {
				line.push(recurringEventRules(item));  //for items with recurrences where we want rules exported
			} else {
				line.push(singleEventRules(item));  //for single items where we want rules exported
			}
		} else {
			if (item.recurrenceInfo) {
				line.push(recurringEventItems(item));  //for items with recurrences where we want occurrences exported
			} else {
				line.push(singleEventItems(item));  //for single items where we want occurrences exported
			}
		}

		str = line.join("");
		if ( prefManager.getBoolPref("export_as_ANSI")) {
			aStream.write(str, str.length);
		} else {
			convStream.writeString(str);
		}
	}

	var footers = [];
	footers.push("END:VCALENDAR");

	str = footers.join("\n")+"\n";
	if ( prefManager.getBoolPref("export_as_ANSI")) {
		aStream.write(str, str.length);
	} else {
			convStream.writeString(str);
	}
	
	return;
}

//Helper functions

function singleEventRules(item) {
	var eventType = "single";
	var line = [];
	var alarmOffset = item.alarmOffset;
	line.push(serializeEvent(item, eventType, alarmOffset));
	details = line.join("");
	return details;
}

function singleEventItems(item) {
	var eventType = "single";
	var alarmOffset = item.alarmOffset;
	var line = [];
	
	if (item.startDate) { date = item.startDate.nativeTime } else if (item.entryDate) { date = item.entryDate.nativeTime } else { date = 0}
	var startDiff = prefManager.getIntPref("export_start_date")*100000000 - date;
	if (item.endDate) { date = item.endDate.nativeTime } else if (item.dueDate) { date = item.dueDate.nativeTime } else { date = 0}
	var endDiff = prefManager.getIntPref("export_end_date")*100000000 - date;
	
	//Get the details about each item between startDate and endDate
	if (startDiff < 0 && endDiff > 0) {
		line.push(serializeEvent(item, eventType, alarmOffset));
	}
	details = line.join("");
	return details;
}

function recurringEventRules(item) {
	var eventType = "single";
	var line = [];
	var alarmOffset = item.alarmOffset;
	line.push(serializeEvent(item, eventType, alarmOffset));
	details = line.join("");
	return details;
}

function recurringEventItems(item) {
	var eventType = "multiple";
	//Define the start time and the end time from the prefs
	var startDate = new Date();
	startDate.setTime(prefManager.getIntPref("export_start_date")*100000);
	startDate = jsDateToDateTime(startDate);
	var endDate = new Date();
	endDate.setTime(prefManager.getIntPref("export_end_date")*100000);
	endDate = jsDateToDateTime(endDate);
	var alarmOffset = item.alarmOffset;
	
	var line = [];
	//Get the details about each occurrence between startDate and endDate
	occurrences = item.getOccurrencesBetween(startDate,endDate,{});

	for each (var occurrence in occurrences) {
		line.push(serializeEvent(occurrence, eventType, alarmOffset));
	}
	details = line.join("");
	return details;
}

function serializeEvent(item, eventType, alarmOffset) {
	var line = [];
	if (item instanceof Components.interfaces.calIEvent) {       
		line.push("BEGIN:VEVENT\n");
		//All day events have only a date attached to them, so if there is no time attached we have to insert one
		//or the resulting file will fail to import into some programs. Hopefully, most programs  will interpret events
		//lasting from midnight to midnight as all day events. We don't use UTC time here because timezones will alter event times.
		if (item.startDate.isDate) {
			//RFC 2445 (iCalendar bible) says all day events have an inclusive start date and a non-inclusive end date.
			//Sunbird/Lightning interprets this as an event goes from day 1 to day 2 with no times attached.
			//vCalendar doesn't have anything in particular to say on the issue, but an event from day 1 at 0:00 to
			//day 1 at 24:00 seems to work for most programs..
			line.push(item.startDate ? "DTSTART:" + item.startDate.icalString + "T000000\n" : "");
			//We start to find problems with poorly formatted ics files which have been imported into Sunbird/Lightning
			//Some put the start date and end date as the same, which Sunbird/Lightning interpret fine, but we have to
			//allow for both situations.  So, if the end date day is the same as the start date day, don't subtract a day.
			//We're probably still screwed for a 2 day event (but that's less common, I hope)
			if (item.endDate.day == item.startDate.day) {
				endDate = item.endDate.day;
			} else {
				endDate = item.endDate.day - 1;
			}
			//Months are stored as from 0 to 11, not 1 to 12
			endMonth = item.startDate.month + 1;
			if (item.endDate) {
				line.push("DTEND:" + item.endDate.year);
				line.push(endMonth < 9 ? "0" + endMonth : endMonth);
				line.push(endDate < 9 ? "0" + endDate + "T240000\n" : endDate + "T240000\n");
			}
		} else {
			line.push(item.startDate ? "DTSTART:" + item.startDate.getInTimezone(UTC()).icalString + "\n" : "");
			line.push(item.endDate ? "DTEND:" + item.endDate.getInTimezone(UTC()).icalString + "\n" : "");
		}
		line.push(item.getProperty("STATUS") ? "STATUS:" + item.getProperty("STATUS") + "\n" : "");

	} else if (item instanceof Components.interfaces.calITodo) {
		//TODOs have a different header on their entries, and also use a different property
		line.push("BEGIN:VTODO\n");
		//We may need to deal with all day tasks (is there such a thing?)
		line.push(item.entryDate ? "DTSTART:" + item.entryDate.getInTimezone(UTC()).icalString + "\n" : "");
		line.push(item.dueDate ? "DUE:" + item.dueDate.getInTimezone(UTC()).icalString + "\n" : "");
		line.push(item.completedDate ? "COMPLETED:" + item.completedDate.getInTimezone(UTC()).icalString + "\n" : "");
		//Some devices require a PRIORITY and STATUS line to be present for tasks.
		line.push(item.getProperty("PRIORITY") ? "PRIORITY:" + item.getProperty("PRIORITY") + "\n" : "PRIORITY:3\n");
		line.push(item.getProperty("STATUS") ? "STATUS:" + item.getProperty("STATUS") + "\n" : "STATUS:ACCEPTED\n");
	}
	
	//We use UID for individual events, and RELATED-TO for occurrences (which end up with no UID at all)
	//Should we use the same UID instead, and the start time as a recurrence Id?  Would that mess things
	//up if we don't have recurrence rules or dates defined in the parent event?
	if (eventType == "single") {
		line.push("UID:" + item.id + "\n");
	} else {
		if (item.recurrenceId == item.parentItem.recurrenceId) {
			line.push("UID:" + item.id + "\n");
		} else {
			line.push("RELATED-TO:" + item.id + "\n");
		}
	}
	
	line.push(item.getProperty("CATEGORIES") ? "CATEGORIES:" + item.getProperty("CATEGORIES") + "\n" : "");
	line.push(item.title ? "SUMMARY:" + item.title + "\n" : "");

	//we need to add the correct line break character code to the description field (=0A=0D)
	//and also, correct the exporting of " characters in Quoted Printable character encoding
	if (item.getProperty("DESCRIPTION")) {
		var description = item.getProperty("DESCRIPTION");
		//We need to keep the line feed at the end of the line as it's not really part of the description
		if (description.indexOf("\n") != "-1"){
			//replace \n with =0D=0A\n (need to keep the line break in the file as well)
			description = description.replace(/\n/g, "=0D=0A\n  ");
		}
	}

	line.push(item.getProperty("DESCRIPTION") ? "DESCRIPTION;ENCODING=QUOTED-PRINTABLE:" + description + "\n" : "");
	line.push(item.getProperty("LOCATION") ? "LOCATION:" + item.getProperty("LOCATION") + "\n" : "");
	line.push(item.getProperty("CLASS") ? "CLASS:" + item.getProperty("CLASS") + "\n" : "");
	line.push(alarmOffset ? "DALARM:" + alarmCalc(alarmOffset, item.startDate.nativeTime) + ";;;\n" : "");
	line.push("DCREATED:" + item.getProperty("CREATED").icalString + "\n");
	line.push("LAST-MODIFIED:" + item.getProperty("LAST-MODIFIED").icalString + "\n");
	line.push(item.getProperty("URL") ? "URL:" + item.getProperty("URL") + "\n" : "");

	//Don't export recurrence rules if recurrences are set to be exported as items
	if (eventType == "single") {
		line.push(item.recurrenceInfo ? "RRULE:" + vcsRecurrenceRule(item) + "\n" : "");
	}

	for each (attendee in item.getAttendees({})) {
		//the properties of attendees for vcal is slightly different than for ical
		line.push("ATTENDEE");
		var role;
		var expect;
		if (attendee.role == "OPT-PARTICIPANT") {role = "ATTENDEE"; expect = "REQUEST"}
		if (attendee.role == "REQ-PARTICIPANT") {role = "ATTENDEE"; expect = "REQUIRE"}
		if (attendee.role == "CHAIR") {role = "ORGANIZER"}
		line.push(attendee.role ? ";ROLE=" + role : "");
		line.push(expect ? ";EXPECT=" + expect : "");
		line.push(attendee.participationStatus ? ";STATUS=" + attendee.participationStatus : "");
		line.push(":" + attendee.id.substring(7) + "\n");
	}

	if (item instanceof Components.interfaces.calIEvent) {
		line.push("END:VEVENT\n");
	} else if (item instanceof Components.interfaces.calITodo) {
		line.push("END:VTODO\n");
	} 
	line = line.join("");
	return line;
}

//Function to calculate alarm times
function alarmCalc(aAlarmOffset, aStartDate) {
	//Get the components of the alarm time as milliseconds.
	var minutes = aAlarmOffset.minutes * 60000;
	var hours = aAlarmOffset.hours * 3600000;
	var days = aAlarmOffset.days * 86400000;
	var weeks = aAlarmOffset.weeks * 604800000;

	//Add them up to get the number of milliseconds the alarm is before the start on the event
	var offset = minutes + hours + days + weeks;
	//Subtract them from the start time of the event
	var alarmTime = aStartDate/1000 - offset;
	//Convert the result into a date format and send it to a function to convert it back to an icalstring in UTC time.
	var date = new Date(alarmTime);
	var alarm = jsDateToDateTime(date);
	return alarm.getInTimezone(UTC()).icalString;
}

function vcsRecurrenceRule(aItem) {
	var recurrenceInfo = aItem.recurrenceInfo;
	var str = [];
	// Get the rules and exceptions
	var ritems = recurrenceInfo.getRecurrenceItems({});
	var rules = [];
	var exceptions = [];
	for each (var r in ritems) {
		if (r.isNegative) {
			exceptions.push(r);
		} else {
			rules.push(r);
		}
	}
	
	if (rules.length > 0) {
		var rule = rules[0];
	}

	const dayTable = {
		1 : "SU",
		2 : "MO",
		3 : "TU",
		4 : "WE",
		5 : "TH",
		6 : "FR",
		7 : "SA"
	};
	if (rule instanceof Components.interfaces.calIRecurrenceRule) {
		// There are four main types of rules, defined by their intervals.
		//Some of them have subtypes as well.
		var count;
		var endDate;
		switch (rule.type) {
			case "DAILY":
				ruleType = "D";
				if (rule.isByCount) {
					if (rule.count == "-1") { count = "0" } else {count = rule.count };
				} else {
					endDate = rule.endDate ? rule.endDate.getInTimezone(UTC()) : null;
				}
				str.push(ruleType + rule.interval + " ");
				str.push(count ? "#" + count : "" );
				str.push(endDate ? endDate.icalString : "");
				break;
			case "WEEKLY":
				ruleType= "W";
				//Weekly recurrences can have a BYDAY component
				if (rule.isByCount) {
					//For infinitely repeating events, the count is 0
					if (rule.count == "-1") { count = "0" } else {count = rule.count };
				} else {
					endDate = rule.endDate ? rule.endDate.getInTimezone(UTC()) : null;
				}
				//Weekly recurrences can have a BYDAY component
				if (rule.getComponent("BYDAY", {})) {
					var chunk = [];
					for each (var i in rule.getComponent("BYDAY", {})) {
						//Correlate each of the days in the BYDAY part of the rule with actual days
						day = dayTable[i];
						chunk.push(day + " ");
					}
					var days = chunk.join('')+"";
				}
				str.push(ruleType + rule.interval + " ");
				str.push(days ? days : "");
				str.push(count ? "#" + count : "" );
				str.push(endDate ? " " + endDate.icalString : "");
				break;
			case "MONTHLY":
				ruleType = "M";
				subType = "";
				if (rule.isByCount) {
					if (rule.count == "-1") { count = "0" } else {count = rule.count };
				} else {
					endDate = rule.endDate ? rule.endDate.getInTimezone(UTC()) : null;
				}
				//Monthly recurrences can have a BYDAY or BYMONTHDAY component
				if (rule.getComponent("BYDAY", {})) { 
					subType = "D";
					var chunk = [];
					for each (var i in rule.getComponent("BYDAY",{})) {
						//We need to find the week of the month and the day of the week here
						//First, see if the component is positive or negative (negative means "from the end of the month")
						if (i < 0) { j = i * -1; k = "-";} else {j = i; k = "+"}
						var weekDay = j % 8;
						dayName = dayTable[weekDay];
						var weekNumber = (j - weekDay ) / 8;
						chunk.push(weekNumber + k + " " + dayName +" ");
					}
					var byDay = chunk.join('')+"";
				}
				if (rule.getComponent("BYMONTHDAY", {})) {
					subType = "P";
					var chunk = [];
					for each (var i in rule.getComponent("BYMONTHDAY",{})) {
						//We can have the last day of the month, a day number
						//or number of days from the end of the month.
						
						if (i == "-1") { 
							chunk.push("LD"); 
						} else if (i < "-1") {
							chunk.push(i + "- ");
						} else if (i > "0") {
							chunk.push(i + " ");
						}
					}
					var byMonthDay = chunk.join('')+"";
				}
				str.push(ruleType + subType + rule.interval + " ");
				str.push(byDay ? byDay : "");
				str.push(byMonthDay ? byMonthDay : "");
				str.push(count ? "#" + count : "" );
				str.push(endDate ? " " + endDate.icalString : "");
				break;
			case "YEARLY":
				ruleType = "Y";
				//There are subrules for years too, but Sunbird and Lightning don't care about them
				if (rule.isByCount) {
					if (rule.count == "-1") { count = "0" } else {count = rule.count };
				} else {
					endDate = rule.endDate ? rule.endDate.getInTimezone(UTC()) : null;
				}
				str.push(ruleType + rule.interval + " ");
				str.push(count ? "#" + count : "" );
				str.push(endDate ? " " + endDate.icalString : "");
				break;
			default:
				//We need some sort of default case
				ruleType = "None";
				break;
		}
	}

	recur = str.join('')+"";
	recur = recur.replace(/  / , " ");
	return recur;
}

//Helper functions duplicated from other files. Prefer to use them direct from other files instead if possible.
function jsDateToDateTime(aDate) {
	var newDate = Components.classes["@mozilla.org/calendar/datetime;1"].createInstance(Components.interfaces.calIDateTime);
	newDate.jsDate = aDate;
	return newDate;
}

function getTimezoneService() {
	if (getTimezoneService.mObject === undefined) {
		getTimezoneService.mObject = Components.classes["@mozilla.org/calendar/timezone-service;1"].getService(Components.interfaces.calITimezoneService);
	}
	return getTimezoneService.mObject;
}

function UTC() {
	if (UTC.mObject === undefined) {
		UTC.mObject = getTimezoneService().UTC;
	}
	return UTC.mObject;
}
