var gBundle = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
var mystrings = gBundle.createBundle("chrome://vcssupport/locale/vcssupport.properties");
var vcsFileType = mystrings.GetStringFromName("vcsFileType");

function calVcsImporter() {
	this.wrappedJSObject = this;
}

calVcsImporter.prototype.QueryInterface = function QueryInterface(aIID) {
	if (!aIID.equals(Components.interfaces.nsISupports) && !aIID.equals(Components.interfaces.calIImporter)) {
		throw Components.results.NS_ERROR_NO_INTERFACE;
	}

	return this;
};

calVcsImporter.prototype.getFileTypes = function getVcsFileTypes(aCount) {
	aCount.value = 1;
	return([{defaultExtension:'vcs', extensionFilter:'*.vcs', description:vcsFileType}]);
};

calVcsImporter.prototype.importFromStream = function openVcs(inputStream, aCount) {
		const nsIScriptableInputStream = Components.interfaces.nsIScriptableInputStream;

		var scriptableInputStream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(nsIScriptableInputStream);

		scriptableInputStream.init(inputStream);
		var stream = scriptableInputStream.read(-1);
		scriptableInputStream.close;
		
		var str = vcsImportReplace(stream);
		var parser = Components.classes["@mozilla.org/calendar/ics-parser;1"].createInstance(Components.interfaces.calIIcsParser);
		parser.parseString(str, null);
		return parser.getItems(aCount);
}

function vcsImportReplace (str) {
	
	var line = [];
	//chop str up into individual lines to help prevent accidentally replacing stuff that we don't want to
	var time = 0;
	while(str.match("\n")) {
		var place = str.indexOf("\n");
		var newstr = str.substring(0,place+1);
		
		//Change the version line to the correct event
		newstr = newstr.replace("VERSION:1.0" , "VERSION:2.0")
		
		//   ***   MS Works specific stuff  ***
		//MS Works uses the wrong version numver in cal files if produces
		newstr = newstr.replace("VERSION:5.0" , "VERSION:2.0");
		//Sunbird and Lightning need all caps in these fields - MS Works uses caps and small letters
		newstr = newstr.replace("BEGIN:vCalendar" , "BEGIN:VCALENDAR");
		newstr = newstr.replace(/BEGIN:vEvent/g , "BEGIN:VEVENT");
		newstr = newstr.replace("END:vCalendar" , "END:VCALENDAR");
		newstr = newstr.replace(/END:vEvent/g , "END:VEVENT");
		
		newstr = newstr.replace("DCREATED:" , "CREATED:");
		if (newstr.match("SUMMARY:")) { var summary = newstr.substring(8);}
		
		newstr = newstr.replace("RRULE: ", "RRULE:");
		
		if (newstr.match("RRULE:")) {
			//Daily Rules
			var interval = newstr.charAt(newstr.indexOf("D") + 1);
			newstr = newstr.replace(/D\d/ , "FREQ=DAILY;INTERVAL=" + interval);
			
			//Weekly Rules
			interval = newstr.charAt(newstr.indexOf("W") + 1);
			newstr = newstr.replace(/W\d/ , "FREQ=WEEKLY;INTERVAL=" + interval);
			
			//Weekly recurrences can have a BYDAY component - we have MO TU etc with spaces
			if (newstr.match("MO" || "TU" ||"WE" ||"TH" || "FR" || "SA" || "SU")) {
				newstr = newstr.replace(/\s/, ";BYDAY=");
			}
			
			//Monthly Rules
			interval = newstr.charAt(newstr.indexOf("M") + 1);
			newstr = newstr.replace(/M\d/ , "FREQ=MONTHLY;INTERVAL=" + interval);

			//Monthly rule can have a BYDAY or BYMONTHDAY component
			if (newstr.match("MP")) {
				interval = newstr.charAt(newstr.indexOf("M") + 2);
				newstr = newstr.replace(/MP\d/, "FREQ=MONTHLY;INTERVAL=" + interval);
				newstr = newstr.replace(/\s/, ";BYMONTHDAY=");
			}
			
			if (newstr.match("MD")) {
				interval = newstr.charAt(newstr.indexOf("M") + 2);
				newstr = newstr.replace(/MD\d/, "FREQ=MONTHLY;INTERVAL=" + interval);
				newstr = newstr.replace(/\s/, ";BYDAY=");
			}
			
			//Yearly Rules
			interval = newstr.charAt(newstr.indexOf("Y") + 1);
			newstr = newstr.replace(/Y\d/, "FREQ=YEARLY;INTERVAL=" + interval);
			
			//Misc Stuff
			
			//An 8 figure number means there is a date.
			var position = newstr.indexOf(newstr.match(/\d\d\d\d\d\d\d\d/));
			var until = newstr.substring(position, position + 8);
			if (str.match(/\d\d\d\d\d\d\d\dT/)) {
				newstr = newstr.replace(/\d\d\d\d\d\d\d\d/, ";UNTIL=" + until);
			} else {
				newstr = newstr.replace(/\d\d\d\d\d\d\d\d/, ";UNTIL=" + until + "T235959Z");
			}
			newstr = newstr.replace(/#/ , ";COUNT=");
			newstr = newstr.replace(/  / , " ");
			newstr = newstr.replace(/; / , ";");
			newstr = newstr.replace(/ ;/ , ";");
			newstr = newstr.replace(/;;/ , ";");
			newstr = newstr.replace(/,,/ , ",");
			newstr = newstr.replace(/ /g , ",");
		}
		
		if (newstr.match("ATTENDEE")) {
			//We need to put attendees stuff here.  Just deleting them for now
			newstr = "";
		}

		//We need the start time to calculate an alarm offset (since for now, Sunbird/Lightning doesn't support absolute alarm times
		if (newstr.match("DTSTART:")) {
			//Get the start time.  It messes up events early in the morning, but it's better than nothing
			var start = newstr.substring(newstr.indexOf(":") + 10, newstr.indexOf(":") + 14)
			hours = start.substring(0,2) * 1;  //Gives us hours
			mins = start.substring(3,4) * 1; //Gives us minutes
			time = (hours * 60) + mins;  //Gives us the time as minutes
		}

		if (newstr.match("DALARM:")) {
			newstr = newstr.replace("\n" , "\nDESCRIPTION:" + summary +"\nEND:VALARM\n");
			//Hack - bound to be unreliable
			var alarmStart = newstr.substring(newstr.indexOf("T") + 1, newstr.indexOf("T") + 5);
			alarmHours = alarmStart.substring(0,2) * 1;
			alarmMins = alarmStart.substring(3,4) * 1;
			alarm = (alarmHours * 60) + alarmMins;
			offset = time - alarm;
			newstr = newstr.replace(/DALARM:\d\d\d\d\d\d\d\dT\d\d\d\d\d\dZ/,"BEGIN:VALARM\nTRIGGER;VALUE=DURATION:-PT" + offset + "M");
			//Preferred method
			//newstr = newstr.replace("DALARM:" , "BEGIN:VALARM\nTRIGGER;VALUE=DATE-TIME:");
			newstr = newstr.replace(";;;" , "")
		}
		

		//Replace the quoted printable code for a new line with a newline character (MS Works uses =0D=0A= rather than =0D=0A for some reason)
		newstr = newstr.replace(/DESCRIPTION;ENCODING=QUOTED-PRINTABLE:=\r\n/ , "DESCRIPTION;ENCODING=QUOTED-PRINTABLE:");
		newstr = newstr.replace(/=0D=0A=\r\n/g, "=0D=0A");
		newstr = newstr.replace(/=0D=0A/g , "\\n");
		newstr = newstr.replace(/,;/ , ";");
		line.push(newstr);
		str=str.substring(place+1);
	}
	//Add the last line in
	line.push(str);
	//Put the lines back together again
	line = line.join("");
	//Replace any extra newlines added with a single newline
	line = line.replace(/\n\n/ , "\n");
	return line;
}
